from unittest import TestCase

import numpy as np
import torch

from dreifus.graphics import homogenize


class GraphicsTest(TestCase):

    def _assert_homogenized(self, points, points_hom):
        for i in range(len(points.shape)):
            if i == len(points.shape) - 1:
                self.assertEqual(points_hom.shape[i], points.shape[i] + 1)
            else:
                self.assertEqual(points_hom.shape[i], points.shape[i])
        self.assertTrue((points == points_hom[..., :-1]).all())
        self.assertTrue((points_hom[..., -1] == 1).all())
        self.assertEqual(points.__class__, points_hom.__class__)

    def test_homogenize(self):
        points = np.random.randn(3, 5, 7, 3)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)

        points = np.random.randn(3, 5, 7, 2)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)

        points = np.random.randn(7, 3)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)

        points = torch.randn(3, 5, 7, 3)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)

        points = torch.randn(3, 5, 7, 2)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)

        points = torch.randn(7, 2)
        points_hom = homogenize(points)

        self._assert_homogenized(points, points_hom)





from math import atan, tan
from typing import Union, Optional

import numpy as np


def fov_to_focal_length(fov: float, size: float) -> float:
    return (size / 2) / (tan(fov / 2))


def focal_length_to_fov(focal_length: float, size: float) -> float:
    return 2 * atan(size / (2 * focal_length))


class Intrinsics(np.ndarray):
    def __new__(cls,
                matrix_or_fx: Union[np.ndarray, float] = np.eye(3),
                fy: Optional[float] = None,
                cx: Optional[float] = None,
                cy: Optional[float] = None,
                s: Optional[float] = None) -> 'Intrinsics':
        intrinsics = super().__new__(cls, (3, 3), dtype=np.float32)
        if not isinstance(matrix_or_fx, np.ndarray) and not isinstance(matrix_or_fx, (float, int)):
            matrix_or_fx = np.asarray(matrix_or_fx)

        if isinstance(matrix_or_fx, np.ndarray) and matrix_or_fx.shape == (3, 3):
            assert fy is None and cx is None and cy is None and s is None, \
                "If a full intrinsics matrix is given, no other parameters should be specified!"
            intrinsics[:] = matrix_or_fx
        elif isinstance(matrix_or_fx, (float, int)):
            assert not (cx is None or cy is None), \
                "If a focal length is given, cx and cy have to be specified!"

            s = 0 if s is None else s
            fy = matrix_or_fx if fy is None else fy

            intrinsics.fill(0)
            intrinsics[0, 0] = matrix_or_fx
            intrinsics[0, 1] = s
            intrinsics[0, 2] = cx
            intrinsics[1, 1] = fy
            intrinsics[1, 2] = cy
            intrinsics[2, 2] = 1
        else:
            print(matrix_or_fx, type(matrix_or_fx))
            raise ValueError("Either a full intrinsics matrix has to be given or fx, cx and cy")

        return intrinsics

    @property
    def fx(self) -> float:
        return self[0, 0].item()

    @property
    def fy(self) -> float:
        return self[1, 1].item()

    @property
    def cx(self) -> float:
        return self[0, 2].item()

    @property
    def cy(self) -> float:
        return self[1, 2].item()

    @property
    def s(self) -> float:
        return self[0, 1].item()

    @fx.setter
    def fx(self, value: float):
        self[0, 0] = value

    @fy.setter
    def fy(self, value: float):
        self[1, 1] = value

    @cx.setter
    def cx(self, value: float):
        self[0, 2] = value

    @cy.setter
    def cy(self, value: float):
        self[1, 2] = value

    @s.setter
    def s(self, value: float):
        self[0, 1] = value

    def get_fovx(self, width: float) -> float:
        return focal_length_to_fov(self.fx, width)

    def get_fovy(self, height: float) -> float:
        return focal_length_to_fov(self.fy, height)


    def rescale(self,
                scale_factor: float,
                scale_factor_y: Optional[float] = None,
                inplace: bool = True) -> 'Intrinsics':
        """
        When images that correspond to this intrinsics matrix are resized, the intrinsics should also be re-scaled
        to account for the image size change.
        This is because the intrinsics effectively just scales
        from the canonical screen space ([-1, 1]^2 with the image center in [0, 0])
        to the image screen space ([0, h] x [0, w] with the image center in [cx, cy]).
        Note: this is an inplace operation.
        See: https://dsp.stackexchange.com/questions/6055/how-does-resizing-an-image-affect-the-intrinsic-camera-matrix

        Parameters
        ----------
            scale_factor:
                scaling factor for resizing the image screen space
            scale_factor_y:
                if specified, the image screen space will be scaled non-uniformly

        Returns
        -------
            The re-scaled intrinsics matrix
        """
        scale_factor_x = scale_factor
        scale_factor_y = scale_factor if scale_factor_y is None else scale_factor_y

        if inplace:
            intrinsics = self
        else:
            intrinsics = self.copy()

        intrinsics[0, 0] *= scale_factor_x  # fx
        intrinsics[1, 1] *= scale_factor_y  # fy
        intrinsics[0, 2] *= scale_factor_x  # cx
        intrinsics[1, 2] *= scale_factor_y  # cy
        # TODO: What about s?

        return intrinsics

    def crop(self,
             crop_left: int = 0,
             crop_top: int = 0,
             inplace: bool = True) -> 'Intrinsics':
        """
        When images that correspond to this intrinsics matrix are cropped, the intrinsics should also be adjusted
        to account for the image size change.
        Essentially, it needs to be ensured that the principal point (in image screen space) after cropping still
        refers to the same point in the canonical screen space.
        This is done by simply subtracting the crop anchor point (top-left) from the intrinsics principal point.
        Note: this is an inplace operation.
        See: https://stackoverflow.com/questions/59477398/how-does-cropping-an-image-affect-camera-calibration-intrinsics

        Examples
        --------
            x: [0, 100] -> cx: 50
            y: [0, 200] -> cy: 100

            crop_x: [20, 90]
            crop_y: [20, 120]
            x: [0, 70]  ([20, 90]) -> cx: (50 - 20) = 30
            y: [0, 100] ([20, 120]) -> cy: (100 - 20) = 80

            => crop() only needs to know crop_left and crop_top

        Parameters
        ----------
            crop_left:
                How many pixels are cropped from left
            crop_top:
                How many pixels are cropped from top

        Returns
        -------
            The intrinsics matrix adjusted for the image cropping operation
        """

        # Ensure that the principal point after cropping is still the same point
        # Only cx, cy are affected

        if inplace:
            intrinsics = self
        else:
            intrinsics = self.copy()

        intrinsics[0, 2] -= crop_left  # cx
        intrinsics[1, 2] -= crop_top  # cy

        return intrinsics

    def homogenize(self, invert: bool = False) -> np.ndarray:
        homogenized = np.eye(4)
        homogenized[:3, :3] = self

        if invert:
            homogenized = np.linalg.inv(homogenized)

        return homogenized

    def invert(self) -> np.ndarray:
        return np.linalg.inv(self)

    def numpy(self) -> np.ndarray:
        return np.array(self)

    def __rmatmul__(self, other):
        if isinstance(other, Intrinsics):
            return super(Intrinsics, self).__rmatmul__(other)
        else:
            return other @ np.array(self)

    def __matmul__(self, other):
        # TODO: figure out why numpy operations automatically cast to Pose again
        if isinstance(other, Intrinsics):
            return super(Intrinsics, self).__matmul__(other)
        else:
            return np.array(self) @ other



import torch


def to_base_tensor(tensor: torch.Tensor):
    # TODO: Not sure if this always works
    tensor.__class__ = torch.Tensor
    return tensor
    # cast = torch.DoubleTensor if tensor.dtype == torch.float64 else torch.FloatTensor
    # return cast(tensor)


from typing import Union, Type
from unittest import TestCase

import numpy as np

from dreifus.matrix.pose_numpy import Pose
from dreifus.matrix.pose_torch import TorchPose
from dreifus.vector.vector_numpy import Vec2, Vec3, Vec4, to_homogeneous
from dreifus.vector.vector_torch import TorchVec4, TorchVec3


class VectorTest(TestCase):

    def _test_vec_2(self, cls: Type[Union[Vec2]]):
        # Empty Vector -> (0, 0)
        vec2 = cls()
        self.assertEqual(vec2.x, 0)
        self.assertEqual(vec2.y, 0)

        vec2 = cls(1, 2)
        self.assertEqual(vec2.x, 1)
        self.assertEqual(vec2.y, 2)

        vec2 = cls([1, 2])
        self.assertEqual(vec2.x, 1)
        self.assertEqual(vec2.y, 2)

        with self.assertRaises(AssertionError):
            vec2 = cls([1, 2], [3, 4])

        with self.assertRaises(ValueError):
            vec2 = cls([1, 2, 3])

        # Currently, specifying fewer arguments than vector length implicitly initalizes the remaining positions with 0
        vec2 = cls(1)
        self.assertEqual(vec2.x, 1)
        self.assertEqual(vec2.y, 0)

        # Vector arithmetics
        vec_a = cls(1, 2)
        vec_b = cls(3, 4)

        vec_c = vec_a + vec_b
        self.assertEqual(vec_c.x, 4)
        self.assertEqual(vec_c.y, 6)

        vec_c = vec_a * vec_b
        self.assertEqual(vec_c.x, 3)
        self.assertEqual(vec_c.y, 8)

        dot = vec_a @ vec_b
        self.assertEqual(dot, 11)

    def _test_vec_3(self, cls: Type[Union[Vec3, TorchVec3]]):
        # Empty Vector -> (0, 0)
        vec3 = cls()
        self.assertEqual(vec3.x, 0)
        self.assertEqual(vec3.y, 0)
        self.assertEqual(vec3.z, 0)

        vec3 = cls(1, 2, 3)
        self.assertEqual(vec3.x, 1)
        self.assertEqual(vec3.y, 2)
        self.assertEqual(vec3.z, 3)

        vec3 = cls([1, 2, 3])
        self.assertEqual(vec3.x, 1)
        self.assertEqual(vec3.y, 2)
        self.assertEqual(vec3.z, 3)

        with self.assertRaises(AssertionError):
            vec3 = cls([1, 2, 3], [3, 4, 5])

        with self.assertRaises(ValueError):
            vec3 = cls([1, 2, 3, 4])

        # Currently, specifying fewer arguments than vector length implicitly initalizes the remaining positions with 0
        vec3 = cls(1)
        self.assertEqual(vec3.x, 1)
        self.assertEqual(vec3.y, 0)
        self.assertEqual(vec3.z, 0)

        # Vector arithmetics
        vec_a = cls(1, 2, 3)
        vec_b = cls(4, 5, 6)

        vec_c = vec_a + vec_b
        self.assertEqual(vec_c.x, 5)
        self.assertEqual(vec_c.y, 7)
        self.assertEqual(vec_c.z, 9)

        vec_c = vec_a * vec_b
        self.assertEqual(vec_c.x, 4)
        self.assertEqual(vec_c.y, 10)
        self.assertEqual(vec_c.z, 18)

        dot = vec_a @ vec_b
        self.assertEqual(dot, 32)

        vec_c = vec_a.cross(vec_b)
        self.assertEqual(vec_c.x, -3)
        self.assertEqual(vec_c.y, 6)
        self.assertEqual(vec_c.x, -3)

        vec_c = vec_a.homogenize()
        self.assertEqual(vec_c.w, 1)

    def _test_vec_4(self, cls: Type[Union[Vec4, TorchVec4]]):
        vec4 = cls(1, 2, 3, 4)
        if cls == TorchVec4:
            pose = TorchPose()
        else:
            pose = Pose()

        vec = pose @ vec4
        self.assertIsInstance(vec, cls)

    def test_vec_2(self):
        self._test_vec_2(Vec2)
        # TODO: There is not TorchVec2 yet

    def test_vec_3(self):
        self._test_vec_3(Vec3)
        self._test_vec_3(TorchVec3)

    def test_vec_4(self):
        self._test_vec_4(Vec4)
        self._test_vec_4(TorchVec4)

    def test_to_homogeneous(self):
        point = Vec3(1, 2, 3)

        point_h = to_homogeneous(point)

        assert np.all(point_h[:3] == np.asarray(point))
        assert point_h[3] == 1
        assert len(point_h.shape) == 1

        points = np.random.randn(100, 3)
        points_h = to_homogeneous(points)

        assert np.all(points_h[..., :3] == np.asarray(points))
        assert np.all(points_h[..., 3] == 1)
        assert len(points_h.shape) == len(points.shape)

        points = np.random.randn(5, 7, 11, 3)
        points_h = to_homogeneous(points)

        assert np.all(points_h[..., :3] == np.asarray(points))
        assert np.all(points_h[..., 3] == 1)
        assert len(points_h.shape) == len(points.shape)



from typing import Optional

import torch

from dreifus.vector.vector_base import Vec3TypeX, unpack_3d_params, unpack_nd_params


def to_homogeneous(points: torch.Tensor) -> torch.Tensor:
    ones = torch.ones((*points.shape[:-1], 1), dtype=points.dtype)
    points = torch.concat([points, ones], dim=-1)

    return points

class TorchVec3(torch.Tensor):
    def __new__(cls,
                x: Optional[Vec3TypeX] = None,
                y: Optional[float] = None,
                z: Optional[float] = None) -> 'TorchVec3':
        x, y, z = unpack_3d_params(x, y, z, default=0)
        vec3 = super().__new__(cls, (x, y, z))

        return vec3

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @x.setter
    def x(self, x: float):
        self[0] = x

    @y.setter
    def y(self, y: float):
        self[1] = y

    @z.setter
    def z(self, z: float):
        self[2] = z

    def length(self):
        return torch.linalg.norm(self)

    def normalize(self) -> 'TorchVec3':
        return self / self.length()

    def homogenize(self) -> 'TorchVec4':
        return TorchVec4(self, 1)

    def tensor(self) -> torch.Tensor:
        return torch.Tensor(self)


class TorchVec4(torch.Tensor):
    def __new__(cls,
                x: Vec3TypeX,
                y: Optional[float] = None,
                z: Optional[float] = None,
                w: Optional[float] = None) -> 'TorchVec4':
        # TODO: Allow no arguments -> 0 vector
        x, y, z, w = unpack_nd_params(4, x, y, z, w)
        vec3 = super().__new__(cls, (x, y, z, w))

        return vec3

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @property
    def w(self) -> float:
        return self[3]

    @x.setter
    def x(self, x: float):
        self[0] = x

    @y.setter
    def y(self, y: float):
        self[1] = y

    @z.setter
    def z(self, z: float):
        self[2] = z

    @w.setter
    def w(self, w: float):
        self[3] = w

    def length(self):
        return torch.linalg.norm(self)

    def normalize(self) -> 'TorchVec4':
        return self / self.length()

    def tensor(self) -> torch.Tensor:
        return torch.Tensor(self)


from typing import Type, Union
from unittest import TestCase

import numpy as np
import torch

from dreifus.matrix.intrinsics_numpy import Intrinsics
from dreifus.matrix.intrinsics_torch import TorchIntrinsics


class IntrinsicsTest(TestCase):

    def _test_intrinsics(self, cls: Type[Union[Intrinsics, TorchIntrinsics]]):
        intrinsics = cls()

        if cls == TorchIntrinsics:
            self.assertTrue((intrinsics == torch.eye(3)).all())
        else:
            self.assertTrue((intrinsics == np.eye(3)).all())

        intrinsics = cls(10, 11, 12, 13)
        self.assertEqual(intrinsics.fx, 10)
        self.assertEqual(intrinsics.fy, 11)
        self.assertEqual(intrinsics.cx, 12)
        self.assertEqual(intrinsics.cy, 13)

        intrinsics_homogenized = intrinsics.homogenize()
        self.assertTrue((intrinsics_homogenized[:3, :3] == intrinsics).all())

        if cls == TorchIntrinsics:
            matrix = torch.arange(3 * 3).reshape(3, 3).float()
        else:
            matrix = np.arange(3 * 3).reshape(3, 3)

        intrinsics_l = intrinsics @ matrix
        intrinsics_r = matrix @ intrinsics

        self.assertNotIsInstance(intrinsics_l, cls)
        self.assertNotIsInstance(intrinsics_r, cls)

    def test_intrinsics(self):
        self._test_intrinsics(Intrinsics)

    def test_torch_intrinsics(self):
        self._test_intrinsics(TorchIntrinsics)


from .vector_base import Vec3Type
from .vector_numpy import Vec2, Vec3, Vec4, rotation_matrix_between_vectors, angle_between_vectors, offset_vector_between_line_and_point
from .vector_torch import TorchVec3, TorchVec4


from unittest import TestCase

from dreifus.camera import CameraCoordinateConvention, AxisDirection, Handedness, PoseType


class CameraTest(TestCase):

    def test_coordinate_convention(self):
        ccc = CameraCoordinateConvention.OPEN_CV
        self.assertEqual(ccc.up_direction, AxisDirection.NEG_Y)
        self.assertEqual(ccc.forward_direction, AxisDirection.Z)
        self.assertEqual(ccc.handedness, Handedness.RIGHT_HANDED)

        ccc = CameraCoordinateConvention.OPEN_GL
        self.assertEqual(ccc.up_direction, AxisDirection.Y)
        self.assertEqual(ccc.forward_direction, AxisDirection.NEG_Z)
        self.assertEqual(ccc.handedness, Handedness.RIGHT_HANDED)

        ccc = CameraCoordinateConvention.DIRECT_X
        self.assertEqual(ccc.up_direction, AxisDirection.Y)
        self.assertEqual(ccc.forward_direction, AxisDirection.Z)
        self.assertEqual(ccc.handedness, Handedness.LEFT_HANDED)

    def test_pose_type(self):
        pose_type = PoseType.CAM_2_WORLD
        self.assertEqual(pose_type.invert(), PoseType.WORLD_2_CAM)
        self.assertEqual(pose_type.invert().invert(), PoseType.CAM_2_WORLD)


from typing import Union, Tuple

import numpy as np

from dreifus.camera import CameraCoordinateConvention, PoseType
from dreifus.matrix import Pose, Intrinsics
from dreifus.vector import Vec3


def back_project(points: np.ndarray, depths: np.ndarray, cam_to_world_pose: Pose, intrinsics: Intrinsics) -> np.ndarray:
    assert len(points.shape) == 2
    assert points.shape[1] == 2
    assert len(depths.shape) == 1
    N = points.shape[0]
    assert depths.shape[0] == N, "Need to have one depth value per point"
    assert cam_to_world_pose.camera_coordinate_convention == CameraCoordinateConvention.OPEN_CV
    assert cam_to_world_pose.pose_type == PoseType.CAM_2_WORLD

    p_screen = np.hstack([points, np.ones((points.shape[0], 1))])
    p_screen_canonical = p_screen @ intrinsics.invert().T
    p_cam = p_screen_canonical * np.expand_dims(depths, 1)
    p_cam_hom = np.hstack([p_cam, np.ones((p_cam.shape[0], 1))])
    p_world = p_cam_hom @ cam_to_world_pose.T

    return p_world[:, :3]


def project(points: np.ndarray, pose: Pose, intrinsics: Intrinsics) -> np.ndarray:
    """
    Projects 3D points onto the image plane defined by the camera pose and intrinsics.

    Parameters
    ----------
        points: 3D points [N, 3]
        pose: Camera pose. Can be either CAM_2_WORLD or WORLD_2_CAM
        intrinsics: Intrinsics of the camera

    Returns
    -------
        Projected points [N, 3] in image space where the third coordinate defines the depth from the camera space.
        Note that no occlusions are handled (difficult for point clouds) and points that would be outside the image
        won't be filtered.
    """

    assert len(points.shape) == 2
    assert points.shape[1] == 3
    cam_to_world_pose = pose.change_pose_type(PoseType.CAM_2_WORLD, inplace=False)
    cam_to_world_pose = cam_to_world_pose.change_camera_coordinate_convention(CameraCoordinateConvention.OPEN_CV,
                                                                              inplace=False)
    world_to_cam_pose = cam_to_world_pose.change_pose_type(PoseType.WORLD_2_CAM, inplace=False)

    p_world = np.hstack([points, np.ones((points.shape[0], 1))])
    p_cam = p_world @ world_to_cam_pose.T
    depths = p_cam[:, [2]]  # Assuming OpenCV convention: depth is positive z-axis
    p_cam = p_cam / depths
    p_screen = p_cam[:, :3] @ intrinsics.T
    p_screen[:, 2] = np.squeeze(depths, 1)  # Return depth as third coordinate

    return p_screen


def draw_onto_image(image: np.ndarray,
                    points: np.ndarray,
                    values: Union[np.ndarray, Tuple]):
    projected_x = points[:, 0].round().astype(int)
    projected_y = points[:, 1].round().astype(int)
    valid_x = (0 <= projected_x) & (projected_x < image.shape[1])
    valid_y = (0 <= projected_y) & (projected_y < image.shape[0])
    valid_xy = valid_x & valid_y

    if isinstance(values, tuple) or isinstance(values, Vec3):
        image[projected_y[valid_xy], projected_x[valid_xy]] = values
    else:
        image[projected_y[valid_xy], projected_x[valid_xy]] = values[valid_xy]



from typing import Union, List, Optional

import numpy as np
from scipy.spatial.transform import Rotation as R
import torch

# ==========================================================
# Torch Matrices
# ==========================================================
from dreifus.matrix.pose_base import is_rotation_matrix
from dreifus.util.typing import to_base_tensor
from dreifus.vector.vector_base import Vec3TypeX, unpack_3d_params, Vec3Type
from dreifus.vector.vector_numpy import Vec4
from dreifus.vector.vector_torch import TorchVec3, TorchVec4


class TorchPose(torch.Tensor):

    def __new__(cls,
                matrix_or_rotation: Union[torch.Tensor, np.ndarray, List] = torch.eye(4),
                translation: Optional[torch.Tensor] = None):
        use_double = isinstance(matrix_or_rotation, torch.Tensor) and matrix_or_rotation.dtype == torch.float64

        def _cast(tensor: torch.Tensor):
            if use_double:
                return tensor.double()
            else:
                return tensor.float()

        if not isinstance(matrix_or_rotation, torch.Tensor):
            matrix_or_rotation = torch.asarray(matrix_or_rotation, dtype=torch.float32)

        if translation is not None and not isinstance(translation, torch.Tensor):
            translation = torch.asarray(translation, dtype=torch.float32)

        device = matrix_or_rotation.device

        if matrix_or_rotation.shape == (4, 4):
            # Full 4x4 Pose
            assert translation is None, "If a full pose is given, no translation should be specified!"
            assert (matrix_or_rotation[3, :] == torch.tensor([0, 0, 0, 1])).all(), \
                f"Last row of pose must be [0, 0, 0, 1]. Got {matrix_or_rotation[3, :]}"
            assert is_rotation_matrix(matrix_or_rotation[:3, :3]), \
                f"Specified matrix does not contain a valid rotation matrix! {matrix_or_rotation[:3, :3]}"

            pose = super().__new__(cls, matrix_or_rotation).to(device)
            pose = _cast(pose)

        elif matrix_or_rotation.shape == (3, 3):
            translation = translation.squeeze()
            # 3x3 rotation matrix + 3(x1) translation vector
            assert translation.shape == (3,), \
                "If a rotation matrix is given, the translation vector has to be 3 dimensional!"

            pose = super().__new__(cls, torch.zeros((4, 4))).to(device)
            pose = _cast(pose)

            pose[:3, :3] = _cast(matrix_or_rotation)
            pose[:3, 3] = _cast(translation)
            pose[3, 3] = 1
        elif matrix_or_rotation.shape == (3, 1) or matrix_or_rotation.shape == (3,):
            # 3(x1) Rodriguez vector + 3(x1) translation vector
            translation = translation.squeeze()
            assert translation.squeeze().shape == (3,), \
                "If a Rodriguez vector is given, the translation vector has to be 3 dimensional!"

            pose = super().__new__(cls, torch.zeros((4, 4))).to(device)
            pose = _cast(pose)

            # TODO: Continue
            pose[:3, :3] = angle_axis_to_rotation_matrix(matrix_or_rotation.unsqueeze(0))[0, :3, :3]
            pose[:3, 3] = _cast(translation)
            pose[3, 3] = 1
        else:
            raise ValueError("Either a full pose has to be given or a 3x3 rotation + 3x1 translation!")

        return pose

    @staticmethod
    def from_euler(
            euler_angles: Vec3Type,
            translation: Vec3Type = TorchVec3(),
            euler_mode: str = 'XYZ') -> 'TorchPose':
        return TorchPose(R.from_euler(euler_mode, euler_angles).as_matrix(), translation)

    @staticmethod
    def from_cv(cv_rotation_vector: TorchVec3, translation: TorchVec3) -> 'TorchPose':
        return TorchPose(cv_rotation_vector, translation)

    def get_rotation_matrix(self) -> torch.Tensor:
        return self[:3, :3]

    # def get_euler_angles(self, order: str) -> TorchVec3:
    #     return Vec3(R.from_matrix(self.get_rotation_matrix()).as_euler(order))

    def get_rodriguez_vector(self) -> TorchVec3:
        return TorchVec3(rotation_matrix_to_angle_axis(self[:3, :].unsqueeze(0))[0])

    # def get_quaternion(self) -> TorchVec4:
    #     return Vec4(R.from_matrix(self.get_rotation_matrix()).as_quat())

    def get_translation(self) -> TorchVec3:
        return TorchVec3(self[:3, 3])

    def set_translation(self, x: Vec3TypeX, y: Optional[float] = None, z: Optional[float] = None):
        x, y, z = unpack_3d_params(x, y, z)
        if x is not None:
            self[0, 3] = x
        if y is not None:
            self[1, 3] = y
        if z is not None:
            self[2, 3] = z

    def invert(self) -> 'TorchPose':
        inverted_rotation = self.get_rotation_matrix().T
        inverted_translation = -inverted_rotation @ self.get_translation()
        inverted_pose = TorchPose(inverted_rotation, inverted_translation)
        return inverted_pose

    def __rmatmul__(self, other):
        if isinstance(other, TorchPose):
            return super(TorchPose, self).__rmatmul__(other)
        else:
            return other @ to_base_tensor(self)

    def __matmul__(self, other):
        if isinstance(other, TorchPose):
            return super(TorchPose, self).__matmul__(other)
        else:
            return to_base_tensor(self) @ other

    def tensor(self) -> torch.Tensor:
        return to_base_tensor(self)


# ==========================================================
# Rotation Helper methods from torchgeometry
# ==========================================================


def angle_axis_to_rotation_matrix(angle_axis):
    """Convert 3d vector of axis-angle rotation to 4x4 rotation matrix

    Args:
        angle_axis (Tensor): tensor of 3d vector of axis-angle rotations.

    Returns:
        Tensor: tensor of 4x4 rotation matrices.

    Shape:
        - Input: :math:`(N, 3)`
        - Output: :math:`(N, 4, 4)`
    """

    def _compute_rotation_matrix(angle_axis, theta2, eps=1e-6):
        # We want to be careful to only evaluate the square root if the
        # norm of the angle_axis vector is greater than zero. Otherwise
        # we get a division by zero.
        k_one = 1.0
        theta = torch.sqrt(theta2)
        wxyz = angle_axis / (theta + eps)
        wx, wy, wz = torch.chunk(wxyz, 3, dim=1)
        cos_theta = torch.cos(theta)
        sin_theta = torch.sin(theta)

        r00 = cos_theta + wx * wx * (k_one - cos_theta)
        r10 = wz * sin_theta + wx * wy * (k_one - cos_theta)
        r20 = -wy * sin_theta + wx * wz * (k_one - cos_theta)
        r01 = wx * wy * (k_one - cos_theta) - wz * sin_theta
        r11 = cos_theta + wy * wy * (k_one - cos_theta)
        r21 = wx * sin_theta + wy * wz * (k_one - cos_theta)
        r02 = wy * sin_theta + wx * wz * (k_one - cos_theta)
        r12 = -wx * sin_theta + wy * wz * (k_one - cos_theta)
        r22 = cos_theta + wz * wz * (k_one - cos_theta)
        rotation_matrix = torch.cat(
            [r00, r01, r02, r10, r11, r12, r20, r21, r22], dim=1)
        return rotation_matrix.view(-1, 3, 3)

    def _compute_rotation_matrix_taylor(angle_axis):
        rx, ry, rz = torch.chunk(angle_axis, 3, dim=1)
        k_one = torch.ones_like(rx)
        rotation_matrix = torch.cat(
            [k_one, -rz, ry, rz, k_one, -rx, -ry, rx, k_one], dim=1)
        return rotation_matrix.view(-1, 3, 3)

    # stolen from ceres/rotation.h

    _angle_axis = torch.unsqueeze(angle_axis, dim=1)
    theta2 = torch.matmul(_angle_axis, _angle_axis.transpose(1, 2))
    theta2 = torch.squeeze(theta2, dim=1)

    # compute rotation matrices
    rotation_matrix_normal = _compute_rotation_matrix(angle_axis, theta2)
    rotation_matrix_taylor = _compute_rotation_matrix_taylor(angle_axis)

    # create mask to handle both cases
    eps = 1e-6
    mask = (theta2 > eps).view(-1, 1, 1).to(theta2.device)
    mask_pos = (mask).type_as(theta2)
    mask_neg = (mask == False).type_as(theta2)  # noqa

    # create output pose matrix
    batch_size = angle_axis.shape[0]
    rotation_matrix = torch.eye(4).to(angle_axis.device).type_as(angle_axis)
    rotation_matrix = rotation_matrix.view(1, 4, 4).repeat(batch_size, 1, 1)
    # fill output matrix with masked values
    rotation_matrix[..., :3, :3] = \
        mask_pos * rotation_matrix_normal + mask_neg * rotation_matrix_taylor
    return rotation_matrix  # Nx4x4


def rotation_matrix_to_angle_axis(rotation_matrix):
    """Convert 3x4 rotation matrix to Rodrigues vector

    Args:
        rotation_matrix (Tensor): rotation matrix.

    Returns:
        Tensor: Rodrigues vector transformation.

    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 3)`
    """
    # todo add check that matrix is a valid rotation matrix
    quaternion = rotation_matrix_to_quaternion(rotation_matrix)
    return quaternion_to_angle_axis(quaternion)


def rotation_matrix_to_quaternion(rotation_matrix, eps=1e-6):
    """Convert 3x4 rotation matrix to 4d quaternion vector

    This algorithm is based on algorithm described in
    https://github.com/KieranWynn/pyquaternion/blob/master/pyquaternion/quaternion.py#L201

    Args:
        rotation_matrix (Tensor): the rotation matrix to convert.

    Return:
        Tensor: the rotation in quaternion

    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 4)`
    """
    if not torch.is_tensor(rotation_matrix):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(rotation_matrix)))

    if len(rotation_matrix.shape) > 3:
        raise ValueError(
            "Input size must be a three dimensional tensor. Got {}".format(
                rotation_matrix.shape))
    if not rotation_matrix.shape[-2:] == (3, 4):
        raise ValueError(
            "Input size must be a N x 3 x 4  tensor. Got {}".format(
                rotation_matrix.shape))

    rmat_t = torch.transpose(rotation_matrix, 1, 2)

    mask_d2 = rmat_t[:, 2, 2] < eps

    mask_d0_d1 = rmat_t[:, 0, 0] > rmat_t[:, 1, 1]
    mask_d0_nd1 = rmat_t[:, 0, 0] < -rmat_t[:, 1, 1]

    t0 = 1 + rmat_t[:, 0, 0] - rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q0 = torch.stack([rmat_t[:, 1, 2] - rmat_t[:, 2, 1],
                      t0, rmat_t[:, 0, 1] + rmat_t[:, 1, 0],
                      rmat_t[:, 2, 0] + rmat_t[:, 0, 2]], -1)
    t0_rep = t0.repeat(4, 1).t()

    t1 = 1 - rmat_t[:, 0, 0] + rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q1 = torch.stack([rmat_t[:, 2, 0] - rmat_t[:, 0, 2],
                      rmat_t[:, 0, 1] + rmat_t[:, 1, 0],
                      t1, rmat_t[:, 1, 2] + rmat_t[:, 2, 1]], -1)
    t1_rep = t1.repeat(4, 1).t()

    t2 = 1 - rmat_t[:, 0, 0] - rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q2 = torch.stack([rmat_t[:, 0, 1] - rmat_t[:, 1, 0],
                      rmat_t[:, 2, 0] + rmat_t[:, 0, 2],
                      rmat_t[:, 1, 2] + rmat_t[:, 2, 1], t2], -1)
    t2_rep = t2.repeat(4, 1).t()

    t3 = 1 + rmat_t[:, 0, 0] + rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q3 = torch.stack([t3, rmat_t[:, 1, 2] - rmat_t[:, 2, 1],
                      rmat_t[:, 2, 0] - rmat_t[:, 0, 2],
                      rmat_t[:, 0, 1] - rmat_t[:, 1, 0]], -1)
    t3_rep = t3.repeat(4, 1).t()

    mask_c0 = mask_d2 * mask_d0_d1
    mask_c1 = mask_d2 * ~mask_d0_d1
    mask_c2 = ~mask_d2 * mask_d0_nd1
    mask_c3 = ~mask_d2 * ~mask_d0_nd1
    mask_c0 = mask_c0.view(-1, 1).type_as(q0)
    mask_c1 = mask_c1.view(-1, 1).type_as(q1)
    mask_c2 = mask_c2.view(-1, 1).type_as(q2)
    mask_c3 = mask_c3.view(-1, 1).type_as(q3)

    q = q0 * mask_c0 + q1 * mask_c1 + q2 * mask_c2 + q3 * mask_c3
    q /= torch.sqrt(t0_rep * mask_c0 + t1_rep * mask_c1 +  # noqa
                    t2_rep * mask_c2 + t3_rep * mask_c3)  # noqa
    q *= 0.5
    return q


def quaternion_to_angle_axis(quaternion: torch.Tensor) -> torch.Tensor:
    """Convert quaternion vector to angle axis of rotation.

    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h

    Args:
        quaternion (torch.Tensor): tensor with quaternions.

    Return:
        torch.Tensor: tensor with angle axis of rotation.

    Shape:
        - Input: :math:`(*, 4)` where `*` means, any number of dimensions
        - Output: :math:`(*, 3)`
    """
    if not torch.is_tensor(quaternion):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(
            type(quaternion)))

    if not quaternion.shape[-1] == 4:
        raise ValueError("Input must be a tensor of shape Nx4 or 4. Got {}"
                         .format(quaternion.shape))
    # unpack input and compute conversion
    q1: torch.Tensor = quaternion[..., 1]
    q2: torch.Tensor = quaternion[..., 2]
    q3: torch.Tensor = quaternion[..., 3]
    sin_squared_theta: torch.Tensor = q1 * q1 + q2 * q2 + q3 * q3

    sin_theta: torch.Tensor = torch.sqrt(sin_squared_theta)
    cos_theta: torch.Tensor = quaternion[..., 0]
    two_theta: torch.Tensor = 2.0 * torch.where(
        cos_theta < 0.0,
        torch.atan2(-sin_theta, -cos_theta),
        torch.atan2(sin_theta, cos_theta))

    k_pos: torch.Tensor = two_theta / sin_theta
    k_neg: torch.Tensor = 2.0 * torch.ones_like(sin_theta)
    k: torch.Tensor = torch.where(sin_squared_theta > 0.0, k_pos, k_neg)

    angle_axis: torch.Tensor = torch.zeros_like(quaternion)[..., :3]
    angle_axis[..., 0] += q1 * k
    angle_axis[..., 1] += q2 * k
    angle_axis[..., 2] += q3 * k
    return angle_axis



from typing import List, Optional, Union, Tuple

import numpy as np
from dreifus.camera import PoseType

from dreifus.matrix import Pose
from dreifus.vector import Vec3, rotation_matrix_between_vectors, offset_vector_between_line_and_point, \
    angle_between_vectors
from dreifus.vector.vector_numpy import rotation_matrix_around_axis


def calculate_look_center(cam_to_world_poses: List[Pose]) -> Vec3:
    a = []  # point on look-direction line (i.e., position of camera)
    d = []  # look direction

    for cam_to_world_pose in cam_to_world_poses:
        position = cam_to_world_pose.get_translation()
        look_direction = cam_to_world_pose.get_look_direction()

        a.append(position)
        d.append(look_direction)

    # See https://stackoverflow.com/questions/48154210/3d-point-closest-to-multiple-lines-in-3d-space
    M = np.zeros((3, 3))
    for k in range(3):
        ms = [d[i][k] * d[i] - (d[i].dot(d[i])) * Vec3.unit(k) for i in range(3)]
        M[k] = sum(ms)

    b = sum([d[i] * (a[i].dot(d[i])) - a[i] * (d[i].dot(d[i])) for i in range(3)])

    look_center = np.linalg.solve(M, b)

    return Vec3(look_center)


def align_poses(world_to_cam_poses: List[Pose],
                up: Optional[Vec3] = Vec3(0, 1, 0),
                look: Optional[Vec3] = Vec3(0, 0, -1),
                look_center: Optional[Vec3] = Vec3(0, 0, 0),
                cam_to_world: bool = False,
                inplace: bool = False,
                return_transformation: bool = False) -> Union[List[Pose], Tuple[List[Pose], Pose]]:
    """
    Calibration poses can be arbitrarily aligned. This method provides a utility to transform a set of camera poses
    such that their up/look directions and look center correspond to the specified values.
    calibration poses are expected in world_to_cam format and OpenCV coordinate convention
    (i.e., x -> right, y -> down, z -> forward).
    Per default, the set of camera poses will be transformed to look at the center in an OpenGL world
    (i.e., x -> right, y -> up, z -> backward).

    Parameters
    ----------
        world_to_cam_poses: the poses to transform
        up: where the up direction should point to
        look: where the look direction should point to
        look_center: where the look center of all cameras should fall into
        cam_to_world: whether the provided poses are already cam_to_world
        return_transformation:
            If specified, a 4x4 transformation matrix is returned that transforms the cam_to_world_poses into
            aligned space exactly as align_poses() would do. Apply as:
                transformation @ world_to_cam_poses[i].invert()

    Returns
    -------
        the re-aligned camera poses
    """

    if not inplace:
        world_to_cam_poses = [pose.copy() for pose in world_to_cam_poses]

    if cam_to_world:
        cam_to_world_poses = world_to_cam_poses
    else:
        cam_to_world_poses = [cam_pose.invert() for cam_pose in world_to_cam_poses]

    transformation = np.eye(4)

    # Align up direction
    if up is not None:
        up_directions = [cam_pose.get_up_direction() for cam_pose in cam_to_world_poses]
        average_up_direction = np.mean(up_directions, axis=0)
        align_up_rotation = rotation_matrix_between_vectors(average_up_direction, up)
        rotator_up = Pose(align_up_rotation, Vec3(), pose_type=PoseType.CAM_2_CAM)
        cam_to_world_poses = [rotator_up @ cam_pose for cam_pose in cam_to_world_poses]
        transformation = rotator_up

    # Align the look direction
    if look is not None:
        look_directions = [cam_pose.get_look_direction() for cam_pose in cam_to_world_poses]
        average_look_direction = np.mean(look_directions, axis=0)
        align_look_rotation = rotation_matrix_between_vectors(average_look_direction, look)
        rotator_look = Pose(align_look_rotation, Vec3(), pose_type=PoseType.CAM_2_CAM)
        cam_to_world_poses = [rotator_look @ cam_pose for cam_pose in cam_to_world_poses]
        transformation = rotator_look @ transformation

    # Align the look center
    if look_center is not None:
        original_look_center = calculate_look_center(cam_to_world_poses)
        offset_vector = look_center - original_look_center

        # look_directions = [cam_pose.get_look_direction() for cam_pose in cam_to_world_poses]
        # cameras_center = np.mean([cam_pose.get_translation() for cam_pose in cam_to_world_poses], axis=0)
        # average_look_direction = np.mean(look_directions, axis=0)
        # # TODO: This won't move cameras much if cameras_center is already at look_center
        # #   Would have to somehow find the point that is closest to all camera rays
        # offset_vector = offset_vector_between_line_and_point(cameras_center, average_look_direction, look_center)
        for cam_pose in cam_to_world_poses:
            cam_pose.move(offset_vector)

        mover = np.eye(4)
        mover[:3, 3] = offset_vector
        transformation = mover @ transformation

    if up is not None:
        # Aligning the look direction might mess up the up direction again
        up_directions = [cam_pose.get_up_direction() for cam_pose in cam_to_world_poses]
        average_up_direction = np.mean(up_directions, axis=0)
        angle = angle_between_vectors(average_up_direction, up)

        # angle_between_vectors() just returns the unsigned angle between the current up direction and the desired up vector
        # We do not know whether we need to rotate clockwise or counter-clockwise
        # The easiest way is to try both rotations and then take the one which aligns better

        # Rotate +angle around look axis
        rotator_1 = Pose(rotation_matrix_around_axis(look, angle), Vec3(), pose_type=PoseType.CAM_2_CAM)
        # rotator = Pose.from_euler(Vec3(0, 0, angle), pose_type=PoseType.CAM_2_CAM)
        cam_to_world_poses_1 = [rotator_1 @ cam_pose for cam_pose in cam_to_world_poses]

        up_directions_1 = [cam_pose.get_up_direction() for cam_pose in cam_to_world_poses_1]
        average_up_direction_1 = np.mean(up_directions_1, axis=0)
        angle_1 = angle_between_vectors(average_up_direction_1, up)

        # Rotate -angle around look axis
        rotator_2 = Pose(rotation_matrix_around_axis(look, -angle), Vec3(), pose_type=PoseType.CAM_2_CAM)
        cam_to_world_poses_2 = [rotator_2 @ cam_pose for cam_pose in cam_to_world_poses]

        up_directions_2 = [cam_pose.get_up_direction() for cam_pose in cam_to_world_poses_2]
        average_up_direction_2 = np.mean(up_directions_2, axis=0)
        angle_2 = angle_between_vectors(average_up_direction_2, up)

        # Figure out which rotation gives better alignment
        if angle_1 < angle_2:
            # Use +angle rotation
            transformation = rotator_1 @ transformation
            cam_to_world_poses = cam_to_world_poses_1
        else:
            # We rotated into the wrong direction. Use -angle rotation
            transformation = rotator_2 @ transformation
            cam_to_world_poses = cam_to_world_poses_2

        up_directions_after_alignment = [cam_pose.get_up_direction() for cam_pose in cam_to_world_poses]
        average_up_direction_after_alignment = np.mean(up_directions_after_alignment, axis=0)
        assert average_up_direction_after_alignment.dot(up) > 0.9, \
            "Up directions could not be properly aligned. This can happen if the desired look direction is something else than the z-axis"

    if return_transformation:
        transformation = Pose(transformation, pose_type=PoseType.CAM_2_CAM)
        return cam_to_world_poses, transformation
    else:
        return cam_to_world_poses



from enum import Enum, auto
from typing import Union, Optional, Tuple

import numpy as np
import torch

ImageType = Union[torch.Tensor, np.ndarray]


class ImageRange(Enum):
    NORMALIZED = (-1, 1)  # [-1, 1]
    FLOAT = (0, 1)  # [0, 1]
    UINT8 = (0, 255)  # [0, 255]

    def min(self) -> int:
        return self.value[0]

    def max(self) -> int:
        return self.value[1]

    def convert_to(self,
                   image: ImageType,
                   target_range: 'ImageRange') -> ImageType:
        src_min = self.min()
        src_max = self.max()
        tgt_min = target_range.min()
        tgt_max = target_range.max()

        return (image - src_min) / (src_max - src_min) * (tgt_max - tgt_min) + tgt_min


class ChannelConvention(Enum):
    CHANNEL_FIRST = auto()  # [C, H, W]
    CHANNEL_LAST = auto()  # [H, W, C]

    def get_permutation(self, target_convention: 'ChannelConvention') -> Optional[Tuple[int, int, int]]:
        if self == target_convention:
            # No conversion necessary
            return None

        if self == ChannelConvention.CHANNEL_FIRST:
            if target_convention == ChannelConvention.CHANNEL_LAST:
                # [C, H, W] -> [H, W, C]
                return 1, 2, 0
        elif self == ChannelConvention.CHANNEL_LAST:
            if target_convention == ChannelConvention.CHANNEL_FIRST:
                # [H, W, C] -> [C, H, W]
                return 2, 0, 1

        raise NotImplementedError(f"Channel Conversion {self} -> {target_convention} not implemented")


class ImageClass(Enum):
    TORCH = auto()
    NUMPY = auto()

    @staticmethod
    def guess(image: ImageType) -> 'ImageClass':
        if isinstance(image, torch.Tensor):
            return ImageClass.TORCH
        elif isinstance(image, np.ndarray):
            return ImageClass.NUMPY
        else:
            raise ValueError(f"Unknown image class: {image.__class__}")

    def permute(self,
                image: ImageType,
                permutation: Tuple[int, int, int]) -> ImageType:
        if self == self.TORCH:
            return image.permute(permutation)
        elif self == self.NUMPY:
            return np.transpose(image, permutation)
        else:
            raise ValueError(f"image class {self} not implemented")

    def convert_to(self,
                   image: ImageType,
                   target_class: 'ImageClass') -> ImageType:
        if self == target_class:
            return image

        if self == self.TORCH:
            if target_class == self.NUMPY:
                return image.detach().cpu().numpy()
        elif self == self.NUMPY:
            if target_class == self.TORCH:
                return torch.tensor(image)

        raise NotImplementedError(f"Image class conversion {self} -> {target_class} not implemented")


class Img:

    def __init__(self,
                 image: ImageType,
                 image_range: ImageRange,
                 channel_convention: ChannelConvention,
                 image_class: Optional[ImageClass] = None
                 ):
        self.img = image
        self.image_range = image_range
        self.channel_convention = channel_convention
        if image_class is None:
            self.image_class = ImageClass.guess(image)
        else:
            self.image_class = image_class

    @staticmethod
    def from_normalized_torch(image: torch.Tensor) -> 'Img':
        # Assume channel first
        return Img(image, ImageRange.NORMALIZED, ChannelConvention.CHANNEL_FIRST, image_class=ImageClass.TORCH)

    @staticmethod
    def from_torch(image: torch.Tensor) -> 'Img':
        # Assume channel first and range [0, 1]
        return Img(image, ImageRange.FLOAT, ChannelConvention.CHANNEL_FIRST, image_class=ImageClass.TORCH)

    @staticmethod
    def from_numpy(image: np.ndarray) -> 'Img':
        if image.dtype in [np.uint8, np.int32]:
            return Img(image, ImageRange.UINT8, ChannelConvention.CHANNEL_LAST, image_class=ImageClass.NUMPY)
        else:
            return Img(image, ImageRange.FLOAT, ChannelConvention.CHANNEL_LAST, image_class=ImageClass.NUMPY)

    def convert_to(self,
                   image_range: Optional[ImageRange] = None,
                   channel_convention: Optional[ChannelConvention] = None,
                   image_class: Optional[ImageClass] = None,
                   inplace: bool = False) -> 'Img':

        image = self.img

        # Image Class
        if image_class is not None:
            image = self.image_class.convert_to(image, image_class)
        else:
            image_class = self.image_class

        # Image Range
        if image_range is not None:
            image = self.image_range.convert_to(image, image_range)
            if image_range == ImageRange.UINT8:
                # Only if target is uint8, we need to explicitly cast
                # In all other cases, the conversion is already done by scaling the range
                if image_class == ImageClass.TORCH:
                    image = image.round().clip(0, 255).type(torch.uint8)
                elif image_class == ImageClass.NUMPY:
                    image = np.clip(image.round(), 0, 255).astype(np.uint8)
                else:
                    raise NotImplementedError(f"uint conversion not implemented for {image_class}")
        else:
            image_range = self.image_range

        # Channel Convention
        if channel_convention is not None:
            permutation = self.channel_convention.get_permutation(channel_convention)
            if permutation is not None:
                image = image_class.permute(image, permutation)
        else:
            channel_convention = self.channel_convention

        if inplace:
            self.img = image
            self.image_range = image_range
            self.channel_convention = channel_convention
            self.image_class = image_class

        return Img(image, image_range=image_range, channel_convention=channel_convention, image_class=image_class)

    def to_numpy(self) -> 'Img':
        # to np.ndarray
        return self.convert_to(
            image_range=ImageRange.UINT8,
            channel_convention=ChannelConvention.CHANNEL_LAST,
            image_class=ImageClass.NUMPY
        )

    def to_torch(self) -> 'Img':
        return self.convert_to(
            image_range=ImageRange.FLOAT,
            channel_convention=ChannelConvention.CHANNEL_FIRST,
            image_class=ImageClass.TORCH
        )

    def to_normalized_torch(self) -> 'Img':
        return self.convert_to(
            image_range=ImageRange.NORMALIZED,
            channel_convention=ChannelConvention.CHANNEL_FIRST,
            image_class=ImageClass.TORCH
        )


def torch_to_numpy_img(torch_img: torch.Tensor) -> np.ndarray:
    return Img.from_torch(torch_img.detach().cpu()).to_numpy().img


def normalized_torch_to_numpy_img(torch_img: torch.Tensor) -> np.ndarray:
    return Img.from_normalized_torch(torch_img.detach().cpu()).to_numpy().img



from unittest import TestCase

import numpy as np

from dreifus.camera_bundle import align_poses
from dreifus.matrix import Pose, Intrinsics
import pyvista as pv
from dreifus.pyvista import add_coordinate_axes, add_camera_frustum


class CameraBundleTest(TestCase):

    def test_align_poses(self):
        camera_1 = Pose.from_euler([np.pi / 4, np.pi / 4, 0], translation=[1, 1, -1])
        camera_2 = Pose.from_euler([np.pi / 4, np.pi / 4, 0], translation=[1, -1, -1])
        camera_3 = Pose.from_euler([np.pi / 4, np.pi / 8, 0], translation=[-1, 1, -1])

        world_to_cam_poses = [camera_1, camera_2, camera_3]
        cam_to_world_poses, transformation = align_poses(world_to_cam_poses, return_transformation=True)

        p = pv.Plotter()
        add_coordinate_axes(p)
        for camera, old_camera in zip(cam_to_world_poses, world_to_cam_poses):

            add_camera_frustum(p, camera, Intrinsics(500, 500, 500, 500))

            cam_to_world_pose = old_camera.invert()
            cam_to_world_pose = transformation @ cam_to_world_pose
            add_camera_frustum(p, cam_to_world_pose, Intrinsics(500, 500, 500, 500), color='red')

            for val_aligned, val_transformed in zip(camera.numpy().flatten(), cam_to_world_pose.numpy().flatten()):
                self.assertAlmostEqual(val_aligned, val_transformed, places=3)
        p.show()


from dataclasses import dataclass
from typing import Union, Tuple, Optional

import numpy as np
import torch
from elias.config import Config

from dreifus.vector.vector_base import unpack_nd_params


@dataclass
class Dimensions(Config, tuple):
    w: int
    h: int

    def __init__(self, w: Union[int, 'Dimensions', Tuple[int, int]], h: Optional[int] = None):
        if isinstance(w, dict):
            # TODO: Is there a better way to match a dictionary against variable names?
            h = w['h']
            w = w['w']

        w, h = unpack_nd_params(2, w, h)
        assert w is not None and h is not None, f"Both w and h have to be given, got: {w}, {h}"

        self.w = w
        self.h = h
        # TODO: Is this a proper dataclass if the super constructor is not called?
        super(Dimensions, self).__init__()  # TODO: Do we need that?

    def __new__(cls, w: Union[int, 'Dimensions', Tuple[int, int]], h: Optional[int] = None):
        if isinstance(w, dict):
            h = w['h']
            w = w['w']

        w, h = unpack_nd_params(2, w, h)
        return tuple.__new__(Dimensions, (w, h))

    # def __getitem__(self, idx: int) -> int:
    #     assert 0 <= idx <= 1, "idx of dimension has to be 0 or 1"
    #
    #     if idx == 0:
    #         return self.w
    #     elif idx == 1:
    #         return self.h

    def __add__(self, other: 'Dimensions') -> 'Dimensions':
        return Dimensions(self.w + other.w, self.h + other.h)

    @property
    def x(self) -> int:
        return self.w

    @property
    def y(self) -> int:
        return self.h


def homogenize(tensor_or_array: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
    """
    Homogenizes the given tensor or array. Works for arbitrary numbers of spatial dimensions.
    Assumes that the spatial dimension is the last in the given array.

    Parameters
    ----------
        tensor_or_array:
            A [..., D] array with arbitrary number of leading dimensions and arbitrary size D of the spatial dimension

    Returns
    -------
        A [..., (D + 1)] array with the last (spatial) dimension extended by ones
    """

    if isinstance(tensor_or_array, np.ndarray):
        ones = np.ones((*tensor_or_array.shape[:-1], 1),
                       dtype=tensor_or_array.dtype)
        return np.concatenate([tensor_or_array, ones], axis=-1)
    elif isinstance(tensor_or_array, torch.Tensor):
        ones = torch.ones((*tensor_or_array.shape[:-1], 1),
                          dtype=tensor_or_array.dtype,
                          device=tensor_or_array.device)
        return torch.concat([tensor_or_array, ones], dim=-1)
    else:
        raise ValueError(f"Unsupported array type: {tensor_or_array.__class__}")


import math
from enum import Enum, auto

import numpy as np

from dreifus.vector import Vec3


class AxisDirection(Enum):
    X = (1, 0, 0)
    NEG_X = (-1, 0, 0)

    Y = (0, 1, 0)
    NEG_Y = (0, -1, 0)

    Z = (0, 0, 1)
    NEG_Z = (0, 0, -1)

    def __init__(self, x: int, y: int, z: int):
        assert sum([x == 0, y == 0, z == 0]) == 2, "Exactly two coordinates have to be zero"
        assert sum([abs(x) == 1, abs(y) == 1, abs(z) == 1]) == 1, "Exactly one coordinate has to be 1 or -1"

        self.axis = Vec3(x, y, z)
        self.axis_id = np.abs(np.array([x, y, z])).argmax()


    def sign(self) -> int:
        return self.axis.sum()

    def axis_name(self) -> str:
        if '_' in self.name:
            axis_name = self.name.split('_')[1]
            axis_name = f"-{axis_name.lower()}"
        else:
            axis_name = self.name.lower()

        return axis_name


class SemanticAxisDirection(Enum):
    RIGHT = (AxisDirection.X)
    LEFT = (AxisDirection.NEG_X)
    UP = (AxisDirection.Y)
    DOWN = (AxisDirection.NEG_Y)
    FORWARD = (AxisDirection.Z)
    BACKWARD = (AxisDirection.NEG_Z)

    def __init__(self, canonical_axis_direction: AxisDirection):
        self.canonical_axis_direction = canonical_axis_direction


class Handedness(Enum):
    RIGHT_HANDED = auto()
    LEFT_HANDED = auto()

    @staticmethod
    def from_sign(sign: int) -> 'Handedness':
        if sign == 1:
            return Handedness.LEFT_HANDED
        elif sign == -1:
            return Handedness.RIGHT_HANDED
        else:
            raise ValueError("Handedness sign must be either 1 or -1")


class Space(Enum):
    CAM = auto()
    WORLD = auto()


class PoseType(Enum):
    CAM_2_WORLD = (Space.CAM, Space.WORLD)
    WORLD_2_CAM = (Space.WORLD, Space.CAM)
    CAM_2_CAM = (Space.CAM, Space.CAM)

    def __init__(self, space_from: Space, space_to: Space):
        self.from_space = space_from
        self.to_space = space_to

    def invert(self) -> 'PoseType':
        if self == PoseType.CAM_2_WORLD:
            return PoseType.WORLD_2_CAM
        elif self == PoseType.WORLD_2_CAM:
            return PoseType.CAM_2_WORLD
        else:
            return PoseType.CAM_2_CAM

    def __matmul__(self, other: 'PoseType') -> 'PoseType':
        assert self.to_space == other.from_space, f"Poses cannot be multiplied: {self.name} and {other.name}"

        new_from_space = self.from_space
        new_to_space = other.to_space

        for pose_type in PoseType:
            if pose_type.from_space == new_from_space and pose_type.to_space == new_to_space:
                return pose_type

        raise ValueError(f"Poses cannot be multiplied: {self.name} and {other.name}. "
                         f"Would result in {new_from_space.name} -> {new_to_space.name}")


class CameraCoordinateConvention(Enum):
    OPEN_CV = (SemanticAxisDirection.RIGHT, SemanticAxisDirection.DOWN, SemanticAxisDirection.FORWARD)
    OPEN_GL = (SemanticAxisDirection.RIGHT, SemanticAxisDirection.UP, SemanticAxisDirection.BACKWARD)  # Also Blender
    DIRECT_X = (SemanticAxisDirection.RIGHT, SemanticAxisDirection.UP, SemanticAxisDirection.FORWARD)  # Also Unity
    PYTORCH_3D = (SemanticAxisDirection.LEFT, SemanticAxisDirection.UP, SemanticAxisDirection.FORWARD)

    def __init__(self,
                 x_direction: SemanticAxisDirection,
                 y_direction: SemanticAxisDirection,
                 z_direction: SemanticAxisDirection):
        self.x_direction = x_direction
        self.y_direction = y_direction
        self.z_direction = z_direction

        # For cameras: Either y or -y is the up direction
        self.up_direction = AxisDirection.Y if y_direction == SemanticAxisDirection.UP else AxisDirection.NEG_Y

        # For cameras: Either z or -z is the forward direction
        self.forward_direction = AxisDirection.Z if z_direction == SemanticAxisDirection.FORWARD else AxisDirection.NEG_Z

        coordinate_system_sign = self.x_direction.canonical_axis_direction.sign() \
                                 * self.y_direction.canonical_axis_direction.sign() \
                                 * self.z_direction.canonical_axis_direction.sign()
        self.handedness = Handedness.from_sign(coordinate_system_sign)

    def __str__(self) -> str:
        string = f"{self.name} ({self.handedness.name}):\n" \
                 f"\t x -> {self.x_direction.name}\n" \
                 f"\t y -> {self.y_direction.name}\n" \
                 f"\t z -> {self.z_direction.name}\n" \
                 f"\t --------------\n" \
                 f"\t up -> {self.up_direction.axis_name()}\n" \
                 f"\t forward -> {self.forward_direction.axis_name()}\n"
        return string


def focal_length_to_fov(focal_length: float, image_size: float):
    """
    Parameters
    ----------
        focal_length:
            focal length (x or y) in pixels or physical millimeters
        image_size:
            width (if focal length defines x) or height of the image plane in pixels or physical millimeters.
            focal_length and image_size have to be both pixels or both millimeters
    """

    return 2 * math.atan(image_size / (2 * focal_length))


if __name__ == '__main__':
    print(CameraCoordinateConvention.OPEN_CV)
    print(CameraCoordinateConvention.DIRECT_X)
    print(CameraCoordinateConvention.OPEN_GL)



from typing import Union, Tuple, List, Optional

import numpy as np
import torch

Vec4Type = Union[Tuple[float, float, float, float], np.ndarray, torch.Tensor, 'Vec4', 'TorchVec4']
Vec4TypeX = Union[float, Vec4Type]  # Can be either the first coordinate ('x') or a full vector
Vec3Type = Union[Tuple[float, float, float], np.ndarray, List[float], torch.Tensor, 'Vec3', 'TorchVec3']
Vec3TypeX = Union[float, Vec3Type]  # Can be either the first coordinate ('x') or a full vector
FloatType = Union[float, np.ndarray]  # Can be single element array (as a result of a numpy operation)


# ==========================================================
# Helper methods for dealing with heterogeneous input types
# ==========================================================

def unpack_nd_params(
        n: int,
        *args,
        default: Optional = None) -> Tuple:
    try:
        if len(args[0]) == n:
            assert all([args[i] is None for i in range(1, n)]), \
                "When a container is passed, all other values must not be given"
            components = args[0]
        elif len(args[0]) == n - 1:
            assert n == 2 or all([args[i] is None for i in range(2, n)]), \
                "When a container is passed, all other values must not be given"
            components = list(args[0]) + [args[1]]
        else:
            raise ValueError(f"Passed container must contain {n} or {n - 1} values!")
    except TypeError:
        # x was actually a number
        components = [default if arg is None else arg for arg in args]
    return components


def unpack_3d_params(
        x: Optional[Vec3TypeX],
        y: float,
        z: float,
        default: Optional = None) -> Tuple[float, float, float]:
    return unpack_nd_params(3, x, y, z, default=default)


def unpack_single(value: FloatType) -> float:
    if isinstance(value, np.ndarray):
        assert value.size == 1, f"Expected array with exactly one element! Got {value.size}"
        return value.item()
    else:
        return value



from typing import Optional, Union

import torch

from dreifus.util.typing import to_base_tensor


class TorchIntrinsics(torch.Tensor):
    def __new__(cls,
                matrix_or_fx: Union[torch.Tensor, float] = torch.eye(3),
                fy: Optional[float] = None,
                cx: Optional[float] = None,
                cy: Optional[float] = None,
                s: Optional[float] = None,
                device: Optional[torch.device] = None,
                dtype: Optional[torch.dtype] = None):

        if not isinstance(matrix_or_fx, torch.Tensor) and not isinstance(matrix_or_fx, (float, int)):
            # Try to pack object into torch Tensor
            matrix_or_fx = torch.as_tensor(matrix_or_fx, device=device, dtype=dtype)

        if isinstance(matrix_or_fx, torch.Tensor) and device is None:
            # Use device of passed tensor
            device = matrix_or_fx.device

        if isinstance(matrix_or_fx, torch.Tensor) and dtype is None:
            # Use dtype of passed tensor
            dtype = matrix_or_fx.dtype

        if isinstance(matrix_or_fx, torch.Tensor) and matrix_or_fx.shape == (3, 3):
            assert fy is None and cx is None and cy is None and s is None, \
                "If a full intrinsics matrix is given, no other parameters should be specified!"
            # Have to use .to() call, as __new__ implicitly calls torch.Tensor which does not support device= keyword
            intrinsics = super().__new__(cls, matrix_or_fx).to(dtype).to(device)

        elif isinstance(matrix_or_fx, (float, int)) or isinstance(matrix_or_fx, torch.Tensor) and matrix_or_fx.shape == (1,):
            assert not (cx is None or cy is None), \
                "If a focal length is given, cx and cy have to be specified!"

            s = 0 if s is None else s
            fy = matrix_or_fx if fy is None else fy

            # Have to use .to() call, as __new__ implicitly calls torch.Tensor which does not support device= keyword
            intrinsics = super().__new__(cls, torch.zeros((3, 3))).to(dtype).to(device)
            if isinstance(matrix_or_fx, torch.Tensor) and matrix_or_fx.dtype == torch.float64:
                intrinsics = intrinsics.double()
            intrinsics[0, 0] = matrix_or_fx
            intrinsics[0, 1] = s
            intrinsics[0, 2] = cx
            intrinsics[1, 1] = fy
            intrinsics[1, 2] = cy
            intrinsics[2, 2] = 1
        else:
            print(matrix_or_fx, type(matrix_or_fx))
            raise ValueError("Either a full intrinsics matrix has to be given or fx, cx and cy")

        return intrinsics

    @property
    def fx(self) -> float:
        return self[0, 0]

    @property
    def fy(self) -> float:
        return self[1, 1]

    @property
    def cx(self) -> float:
        return self[0, 2]

    @property
    def cy(self) -> float:
        return self[1, 2]

    @property
    def s(self) -> float:
        return self[0, 1]

    def homogenize(self) -> torch.Tensor:
        homogenized = torch.eye(4)
        homogenized[:3, :3] = self
        return homogenized

    def tensor(self) -> torch.Tensor:
        return to_base_tensor(self)

    def __rmatmul__(self, other):
        assert not isinstance(other, TorchIntrinsics), "Matrix Multiplication between intrinsics does not make sense"
        return other @ to_base_tensor(self)

    def __matmul__(self, other):
        assert not isinstance(other, TorchIntrinsics), "Matrix Multiplication between intrinsics does not make sense"
        return to_base_tensor(self) @ other


import torch


def _scalar_triple_product(a: torch.Tensor, b: torch.Tensor, c: torch.Tensor) -> torch.Tensor:
    return torch.bmm(a.view(-1, 1, 3), torch.cross(b, c).view(-1, 3, 1)).squeeze(2).squeeze(1)


def _compute_tetrahedron_barycentric_coordinates(a: torch.Tensor,
                                                 b: torch.Tensor,
                                                 c: torch.Tensor,
                                                 d: torch.Tensor,
                                                 p: torch.Tensor) -> torch.Tensor:
    vap = p - a
    vbp = p - b
    vab = b - a
    vac = c - a
    vad = d - a
    vbc = c - b
    vbd = d - b

    va6 = _scalar_triple_product(vbp, vbd, vbc)
    vb6 = _scalar_triple_product(vap, vac, vad)
    vc6 = _scalar_triple_product(vap, vad, vab)
    vd6 = _scalar_triple_product(vap, vab, vac)

    v6 = 1 / _scalar_triple_product(vab, vac, vad)
    return torch.stack([va6 * v6, vb6 * v6, vc6 * v6, vd6 * v6], dim=-1).to(p)


def apply_scene_flow_colormap(offsets: torch.Tensor) -> torch.Tensor:
    """
    Colorizes the given offsets as follows:
     -  0, 0, 0 -> white
     -  1, 0, 0 / 0, 1, 0 / 0, 0, 1 -> red/green/blue
     - -1, 0, 0 / 0,-1, 0 / 0, 0,-1 -> cyan/purple/orange

    This is done by subdividing the 3D xyz offset space into 8 tetrahedrons (each contains the origin and two axis points)
    The corner points of the tetrahedrons are the above-mentioned colors and the color interpolation is found via
    barycentric coordinates of the offset point within its containing tetrahedron

    Args:
        offsets: The offsets to colorize

    Taken from: https://stackoverflow.com/questions/38545520/barycentric-coordinates-of-a-tetrahedron

    Returns:
        colors for the given offsets
    """

    points = torch.tensor([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [-1, 0, 0],
        [0, -1, 0],
        [0, 0, -1]], dtype=torch.float32).to(offsets)

    colors = torch.tensor([
        [1, 1, 1],
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [0, 1, 1],
        [1, 0, 1],
        [1, 1, 0]
    ], dtype=torch.float32).to(offsets)

    original_shape = offsets.shape
    offsets = offsets.view(-1, 3)  # Ensure we just have a list of points

    # TODO: Need better way to choose containing/nearest tetrahedron. It can happen that chosen 4 points are not a
    #   Tetrahedron because 2 of them are opposite to each other!
    distances = (points[None, ...] - offsets[:, None, :]).norm(dim=2)  # [B, 7]
    idx_sorted = torch.sort(distances, dim=1).indices  # [B, 7]
    closest_4_points = points[idx_sorted[:, :4]]

    barycentric_coordinates = _compute_tetrahedron_barycentric_coordinates(closest_4_points[:, 0],
                                                                           closest_4_points[:, 1],
                                                                           closest_4_points[:, 2],
                                                                           closest_4_points[:, 3],
                                                                           offsets)

    barycentric_coordinates_clipped = torch.clip(barycentric_coordinates, min=0, max=1)
    barycentric_coordinates_clipped /= barycentric_coordinates_clipped.sum(dim=1).view(-1, 1)

    interpolated_colors = torch.bmm(barycentric_coordinates_clipped.view(-1, 1, 4), colors[idx_sorted[:, :4]]).squeeze(
        1)
    interpolated_colors = torch.clip(interpolated_colors, min=0, max=1)

    # It can happen that the barycentric coordinates are nan. This is the case if an offset is exactly 0.
    # Then it is not possible to decide for the correct tetrahedron by just the 4 nearest vertices.
    # Since these are rare edge cases, just use white color.
    interpolated_colors[barycentric_coordinates.isnan().any(dim=1)] = colors[0]
    interpolated_colors = interpolated_colors.view(original_shape)

    return interpolated_colors



from unittest import TestCase

import numpy as np
import torch

from dreifus.image import Img, ImageClass, ImageRange, ChannelConvention


class ImageTest(TestCase):

    def _get_numpy_image(self) -> np.ndarray:
        return np.arange(5 * 7 * 3).reshape((5, 7, 3)).astype(np.uint8)

    def _get_torch_image(self) -> torch.Tensor:
        return torch.arange(5 * 7 * 3).reshape((3, 5, 7)) / 255.

    def _get_normalized_torch_image(self) -> torch.Tensor:
        return torch.arange(5 * 7 * 3).reshape((3, 5, 7)) / 255. * 2 - 1

    def test_cast(self):
        # numpy -> torch
        numpy_image = self._get_numpy_image()
        image = Img.from_numpy(numpy_image)
        image = image.convert_to(image_class=ImageClass.TORCH)

        assert isinstance(image.img, torch.Tensor)
        assert image.image_class == ImageClass.TORCH
        assert image.image_range == ImageRange.UINT8
        assert image.channel_convention == ChannelConvention.CHANNEL_LAST
        assert np.all(image.img.numpy() == numpy_image)

        # torch -> numpy
        torch_image = self._get_torch_image()
        image = Img.from_torch(torch_image)
        image = image.convert_to(image_class=ImageClass.NUMPY)

        assert isinstance(image.img, np.ndarray)
        assert image.image_class == ImageClass.NUMPY
        assert image.image_range == ImageRange.FLOAT
        assert image.channel_convention == ChannelConvention.CHANNEL_FIRST
        assert np.all(image.img == torch_image.numpy())

    def test_range_conversion(self):
        # [0, 255] -> [0, 1]
        numpy_image = self._get_numpy_image()
        imageV2 = Img.from_numpy(numpy_image)
        image = imageV2.convert_to(image_range=ImageRange.FLOAT)

        assert np.all(image.img == numpy_image / 255.)
        assert image.img.min() >= 0
        assert image.img.max() <= 1

        # [0, 255] -> [-1, 1]
        image = imageV2.convert_to(image_range=ImageRange.NORMALIZED)

        assert np.all(image.img == (numpy_image / 255. * 2 - 1))
        assert image.img.min() >= -1
        assert image.img.max() <= 1

        # [0, 1] -> [0, 255]
        torch_image = self._get_torch_image()
        imageV2 = Img.from_torch(torch_image)
        image = imageV2.convert_to(image_range=ImageRange.UINT8)

        assert (image.img == torch_image * 255).all()
        assert image.img.min() >= 0
        assert image.img.max() <= 255
        assert image.img.dtype == torch.uint8

        # [0, 1] -> [-1, 1]
        image = imageV2.convert_to(image_range=ImageRange.NORMALIZED)

        assert (image.img == torch_image * 2 - 1).all()
        assert image.img.min() >= -1
        assert image.img.max() <= 1

        # [-1, 1] -> [0, 255]
        normalized_torch_image = self._get_normalized_torch_image()
        imageV2 = Img.from_normalized_torch(normalized_torch_image)
        image = imageV2.convert_to(image_range=ImageRange.UINT8)

        assert (image.img / 255. * 2 - 1 == normalized_torch_image).all()
        assert image.img.min() >= 0
        assert image.img.max() <= 255
        assert image.img.dtype == torch.uint8

    def test_channel_conversion(self):
        # CHANNEL_LAST -> CHANNEL_FIRST

        numpy_image = self._get_numpy_image()
        imageV2 = Img.from_numpy(numpy_image)
        image = imageV2.convert_to(channel_convention=ChannelConvention.CHANNEL_FIRST)

        assert image.img.shape[0] == numpy_image.shape[2]

        # CHANNEL_FIRST -> CHANNEL_LAST

        torch_image = self._get_torch_image()
        imageV2 = Img.from_torch(torch_image)
        image = imageV2.convert_to(channel_convention=ChannelConvention.CHANNEL_LAST)

        assert image.img.shape[2] == torch_image.shape[0]

    def test_full_conversions(self):
        # NUMPY -> TORCH
        numpy_image = self._get_numpy_image()
        imageV2 = Img.from_numpy(numpy_image)
        image_numpy_to_torch = imageV2.to_torch()

        assert image_numpy_to_torch.img.min() >= 0
        assert image_numpy_to_torch.img.max() <= 1
        assert image_numpy_to_torch.img.shape[0] == numpy_image.shape[2]

        # NUMPY -> TORCH_NORMALIZED
        image_numpy_to_torch_normalized = imageV2.to_normalized_torch()

        assert image_numpy_to_torch_normalized.img.min() >= -1
        assert image_numpy_to_torch_normalized.img.max() <= 1
        assert image_numpy_to_torch_normalized.img.shape[0] == numpy_image.shape[2]

        # TORCH -> NUMPY
        torch_image = self._get_torch_image()
        imageV2 = Img.from_torch(torch_image)
        image_torch_to_numpy = imageV2.to_numpy()

        assert image_torch_to_numpy.img.min() >= 0
        assert image_torch_to_numpy.img.max() <= 255
        assert image_torch_to_numpy.img.shape[2] == torch_image.shape[0]
        assert image_torch_to_numpy.img.dtype == np.uint8

        # TORCH -> TORCH_NORMALIZED
        image_torch_to_torch_normalized = imageV2.to_normalized_torch()

        assert image_torch_to_torch_normalized.img.min() >= -1
        assert image_torch_to_torch_normalized.img.max() <= 1
        assert image_torch_to_torch_normalized.img.shape == torch_image.shape

        # TORCH_NORMALIZED -> NUMPY
        torch_normalized_image = self._get_normalized_torch_image()
        imageV2 = Img.from_normalized_torch(torch_normalized_image)
        image_torch_normalized_to_numpy = imageV2.to_numpy()

        assert image_torch_normalized_to_numpy.img.min() >= 0
        assert image_torch_normalized_to_numpy.img.max() <= 255
        assert image_torch_normalized_to_numpy.img.shape[2] == torch_image.shape[0]
        assert image_torch_normalized_to_numpy.img.dtype == np.uint8

        # TORCH_NORMALIZED -> TORCH
        image_torch_normalized_to_torch = imageV2.to_torch()

        assert image_torch_normalized_to_torch.img.min() >= 0
        assert image_torch_normalized_to_torch.img.max() <= 1
        assert image_torch_normalized_to_torch.img.shape == torch_normalized_image.shape

        # NUMPY cycle
        image_cycle_numpy = image_numpy_to_torch.to_numpy()
        assert image_cycle_numpy.image_class == image_torch_normalized_to_numpy.image_class
        assert image_cycle_numpy.image_range == image_torch_normalized_to_numpy.image_range
        assert image_cycle_numpy.channel_convention == image_torch_normalized_to_numpy.channel_convention
        assert np.all(image_cycle_numpy.img == numpy_image)

        # TORCH cycle
        image_cycle_torch = image_torch_to_numpy.to_torch()
        assert image_cycle_torch.image_class == image_torch_normalized_to_torch.image_class
        assert image_cycle_torch.image_range == image_torch_normalized_to_torch.image_range
        assert image_cycle_torch.channel_convention == image_torch_normalized_to_torch.channel_convention
        assert (image_cycle_torch.img == torch_image).all()

        # TORCH_NORMALIZED cycle
        image_cycle_torch_normalized = image_torch_normalized_to_numpy.to_normalized_torch()
        assert image_cycle_torch_normalized.image_class == image_torch_to_torch_normalized.image_class
        assert image_cycle_torch_normalized.image_range == image_torch_to_torch_normalized.image_range
        assert image_cycle_torch_normalized.channel_convention == image_torch_to_torch_normalized.channel_convention
        assert (image_cycle_torch_normalized.img == torch_normalized_image).all()

    def test_guess_image_class(self):
        numpy_image = self._get_numpy_image()
        image = Img(numpy_image, ImageRange.UINT8, ChannelConvention.CHANNEL_LAST)
        assert image.image_class == ImageClass.NUMPY

        torch_image = self._get_torch_image()
        image = Img(torch_image, ImageRange.FLOAT, ChannelConvention.CHANNEL_FIRST)
        assert image.image_class == ImageClass.TORCH



from typing import Union, Type
from unittest import TestCase

import numpy as np
import torch
from numpy.ma.testutils import assert_array_almost_equal, assert_array_equal
from scipy.spatial.transform import Rotation as R

from dreifus.camera import CameraCoordinateConvention, PoseType
from dreifus.matrix.intrinsics_numpy import Intrinsics
from dreifus.matrix.intrinsics_torch import TorchIntrinsics
from dreifus.matrix.pose_numpy import Pose
from dreifus.matrix.pose_torch import TorchPose


class PoseTest(TestCase):

    def _test_pose(self, cls: Type[Union[Pose, TorchPose]]):
        pose = cls()
        self.assertEqual(pose[0, 0], 1)
        self.assertEqual(pose[1, 1], 1)
        self.assertEqual(pose[2, 2], 1)
        self.assertEqual(pose[3, 3], 1)

        rotation_matrix = R.from_euler('xyz', [np.pi, 0, 0]).as_matrix()
        pose = cls(rotation_matrix, [1, 2, 3])

        translation = pose.get_translation()
        assert_array_equal(translation, [1, 2, 3])

        euler_angles = pose.get_euler_angles('xyz')
        assert_array_almost_equal(euler_angles, [np.pi, 0, 0])

        pose = cls.from_euler([np.pi, 0, 0])
        euler_angles = pose.get_euler_angles('xyz')
        assert_array_almost_equal(euler_angles, [np.pi, 0, 0])

        rodriguez = R.from_euler('xyz', [np.pi, 0, 0]).as_rotvec()
        pose = cls.from_rodriguez(rodriguez)
        euler_angles = pose.get_euler_angles('xyz')
        assert_array_almost_equal(euler_angles, [np.pi, 0, 0])

        pose = pose.invert()
        euler_angles = pose.get_euler_angles('xyz')
        assert_array_almost_equal(euler_angles, [-np.pi, 0, 0])

        if cls == TorchPose:
            matrix = torch.arange(4 * 4).reshape(4, 4).float()
        else:
            matrix = np.arange(4 * 4).reshape(4, 4)

        pose_l = pose @ matrix
        pose_r = matrix @ pose

        self.assertNotIsInstance(pose_l, cls)
        self.assertNotIsInstance(pose_r, cls)

        pose_other = cls()

        # pose is CAM_2_WORLD due to invert()
        pose_l = pose @ pose_other
        pose_r = pose_other.invert() @ pose.invert()

        with self.assertRaises(ValueError):
            # Should not work as pose_other is WORLD_2_CAM and pose is CAM_2_WORLD. WORLD -> WORLD doesn't exist
            pose_other @ pose

        self.assertIsInstance(pose_l, cls)
        self.assertIsInstance(pose_r, cls)

        if cls == TorchPose:
            intrinsics = TorchIntrinsics()
        else:
            intrinsics = Intrinsics()

        pose_l = pose @ intrinsics.homogenize()
        pose_r = intrinsics.homogenize() @ pose

        self.assertNotIsInstance(pose_l, cls)
        self.assertNotIsInstance(pose_r, cls)

        pose = Pose(camera_coordinate_convention=CameraCoordinateConvention.DIRECT_X, pose_type=PoseType.CAM_2_WORLD)
        pose_copy = pose.copy()

        self.assertEqual(pose.camera_coordinate_convention, pose_copy.camera_coordinate_convention)
        self.assertEqual(pose.pose_type, pose_copy.pose_type)

        pose_change = pose.change_camera_coordinate_convention(CameraCoordinateConvention.OPEN_CV)
        # Changing the camera coordinate convention should affect some of the entries in the matrix
        self.assertFalse((pose == pose_change).all())

    def test_pose(self):
        self._test_pose(Pose)

    def test_torch_pose(self):
        self._test_pose(TorchPose)



from unittest import TestCase

import pyvista as pv
from elias.util import load_img
from elias.util.io import resize_img
from pyvista.examples import download_mars_jpg

from dreifus.matrix import Pose, Intrinsics
from dreifus.pyvista import add_camera_frustum, add_coordinate_axes


class PyVistaTest(TestCase):

    def test_add_camera_frustum(self):
        img_mars = resize_img(load_img(download_mars_jpg()), 1 / 4)
        W = img_mars.shape[1]
        H = img_mars.shape[0]

        pose = Pose()
        intrinsics = Intrinsics(1000, 1000, W / 2, H / 2)

        p = pv.Plotter()
        add_coordinate_axes(p, scale=0.1)
        add_camera_frustum(p, pose, intrinsics, image=img_mars / 255)
        p.show()



from math import cos, sin
from typing import Optional, List

import numpy as np

from dreifus.camera import PoseType
from dreifus.matrix import Pose
from dreifus.vector import Vec3


def point_around_axis(theta: float,
                      axis: Vec3 = Vec3(0, 0, 1)) -> Vec3:
    """
    Compute a point with unit distance from `axis` that is rotated by `theta` around it.
    It is somewhat arbitrary where `theta=0` lands.
     - (1, 0, 0) -> (0, 0, 1)
     - (0, 1, 0) -> (0, 0, -1)
     - (0, 0, 1) -> (0, 1, 0)

    Computed points via `point_around_axis()` will be centered around the origin.

    Parameters
    ----------
        theta: angle between [0, 2pi) specifying the rotation around the axis
        axis: the axis to rotate around
    """

    axis = Vec3(axis)
    non_parallel = Vec3(0, 1, 0) if (axis == Vec3(1, 0, 0)).all() else Vec3(1, 0, 0)
    v = axis.cross(non_parallel).normalize()
    v_rotated = cos(theta) * v + sin(theta) * (np.cross(axis, v)) + (1 - cos(theta)) * axis * (np.dot(axis, v))

    return v_rotated


def circle_around_axis(n_poses: int,
                       axis=Vec3(0, 0, 1),
                       up: Vec3 = Vec3(0, 0, 1),
                       move=Vec3(),
                       distance: float = 1.0,
                       distance_end: Optional[float] = None,
                       theta_from: float = 0,
                       theta_to: float = 2 * np.pi,
                       look_at: Vec3 = Vec3(0, 0, 0)) -> List[Pose]:
    """
    Computes `n_poses` many camera poses (cam2world) that circle with distance `distance` around the specified `axis`
    that is moved by `move`.
    Per default, one full circle is computed. With `theta_from` and `theta_to` one can specify parts of the circle
    or even multiple circulations around the axis.

    Parameters
    ----------
        n_poses:
            how many poses should be computed
        axis:
            The axis (direction) around which we rotate
        up:
            which direction should be up for the camera
        move:
            the location of the axis
        distance:
            distance of the pose locations from the axis
        distance_end:
            if specified, `distance` will be interpreted as a start distance for the first pose and distance_end
            defines the distance from the axis for the last pose. Distances in for poses in between are linearly
            interpolated. This gives a spiraling effect.
        theta_from:
            orientation of the first pose
        theta_to:
            orientation of the last pose
        look_at:
            all circle poses will look at the specified point in space. Per default, this is the origin
    """

    if distance_end is None:
        distance_end = distance
        distance_start = distance
    else:
        distance_start = distance

    poses = []
    for i_pose in range(n_poses):
        alpha = i_pose / n_poses
        theta = theta_from + alpha * (theta_from - theta_to)
        distance = distance_start + alpha * (distance_end - distance_start)
        location = distance * point_around_axis(theta, axis=axis)

        pose = Pose(pose_type=PoseType.CAM_2_WORLD)
        location += move
        pose.set_translation(location)
        pose.look_at(look_at, up=up)

        poses.append(pose)

    return poses



from typing import Union

import numpy as np
import torch


def is_rotation_matrix(matrix: np.ndarray):
    # Taken from https://stackoverflow.com/questions/53808503/how-to-test-if-a-matrix-is-a-rotation-matrix
    I = np.identity(matrix.shape[0])
    return np.allclose((np.matmul(matrix, matrix.T)), I, atol=1e-5) and (
        np.isclose(np.linalg.det(matrix), 1, atol=1e-8))

def ensure_homogenized(vector_or_matrix: Union[torch.Tensor, np.ndarray]):
    if isinstance(vector_or_matrix, torch.Tensor):
        if len(vector_or_matrix) == 1 and vector_or_matrix.shape[0] == 3:
            # (3,) Vector
            vector_or_matrix = torch.concat([vector_or_matrix, torch.ones((1,))])

        elif vector_or_matrix.shape[1] == 3:
            # N x 3 Matrix
            vector_or_matrix = torch.concat([vector_or_matrix, torch.ones((vector_or_matrix.shape[0], 1))], dim=1)
    else:
        if len(vector_or_matrix) == 1 and vector_or_matrix.shape[0] == 3:
            # (3,) Vector
            vector_or_matrix = np.concatenate([vector_or_matrix, np.ones((1,))])

        elif vector_or_matrix.shape[1] == 3:
            # N x 3 Matrix
            vector_or_matrix = np.concatenate([vector_or_matrix, np.ones((vector_or_matrix.shape[0], 1))], axis=1)

    return vector_or_matrix


import numpy as np
from trimesh import grouping
from trimesh.geometry import faces_to_edges


def subdivide(vertices,
              faces,
              face_index=None,
              vertex_attributes=None):
    """
    Adapted from trimesh, because trimesh apparently treats vertex_attributes wrongly in subdivision.

    Subdivide a mesh into smaller triangles.

    Note that if `face_index` is passed, only those
    faces will be subdivided and their neighbors won't
    be modified making the mesh no longer "watertight."

    Parameters
    ------------
    vertices : (n, 3) float
      Vertices in space
    faces : (m, 3) int
      Indexes of vertices which make up triangular faces
    face_index : faces to subdivide.
      if None: all faces of mesh will be subdivided
      if (n,) int array of indices: only specified faces
    vertex_attributes : dict
      Contains (n, d) attribute data
    return_index : bool
      If True, return index of original face for new faces

    Returns
    ----------
    new_vertices : (q, 3) float
      Vertices in space
    new_faces : (p, 3) int
      Remeshed faces
    index_dict : dict
      Only returned if `return_index`, {index of
      original face : index of new faces}.
    """
    if face_index is None:
        face_mask = np.ones(len(faces), dtype=bool)
    else:
        face_mask = np.zeros(len(faces), dtype=bool)
        face_mask[face_index] = True

    # the (c, 3) int array of vertex indices
    faces_subset = faces[face_mask]

    # find the unique edges of our faces subset
    edges = np.sort(faces_to_edges(faces_subset), axis=1)
    unique, inverse = grouping.unique_rows(edges)
    # then only produce one midpoint per unique edge
    mid = vertices[edges[unique]].mean(axis=1)
    mid_idx = inverse.reshape((-1, 3)) + len(vertices)

    # the new faces_subset with correct winding
    f = np.column_stack([faces_subset[:, 0],
                         mid_idx[:, 0],
                         mid_idx[:, 2],
                         mid_idx[:, 0],
                         faces_subset[:, 1],
                         mid_idx[:, 1],
                         mid_idx[:, 2],
                         mid_idx[:, 1],
                         faces_subset[:, 2],
                         mid_idx[:, 0],
                         mid_idx[:, 1],
                         mid_idx[:, 2]]).reshape((-1, 3))

    # add the 3 new faces_subset per old face all on the end
    # by putting all the new faces after all the old faces
    # it makes it easier to understand the indexes
    new_faces = np.vstack((faces[~face_mask], f))
    # stack the new midpoint vertices on the end
    new_vertices = np.vstack((vertices, mid))

    if vertex_attributes is not None:
        new_attributes = {}
        for key, values in vertex_attributes.items():
            interpolated_attributes = values[edges[unique]].mean(axis=1)
            new_attributes[key] = np.vstack((values, interpolated_attributes))
        return new_vertices, new_faces, new_attributes

    return new_vertices, new_faces


# TODO: Implement generic Transform class: 4x4 matrix with scale
from typing import Tuple

import numpy as np


def compute_similarity_transform(from_points: np.ndarray, to_points: np.ndarray) \
        -> Tuple[float, np.ndarray, np.ndarray]:
    assert len(from_points.shape) == 2, \
        "from_points must be a m x n array"
    assert from_points.shape == to_points.shape, \
        "from_points and to_points must have the same shape"

    N, m = from_points.shape

    mean_from = from_points.mean(axis=0)
    mean_to = to_points.mean(axis=0)

    delta_from = from_points - mean_from  # N x m
    delta_to = to_points - mean_to  # N x m

    sigma_from = (delta_from * delta_from).sum(axis=1).mean()
    sigma_to = (delta_to * delta_to).sum(axis=1).mean()

    cov_matrix = delta_to.T.dot(delta_from) / N

    U, d, V_t = np.linalg.svd(cov_matrix, full_matrices=True)
    cov_rank = np.linalg.matrix_rank(cov_matrix)
    S = np.eye(m)

    if cov_rank >= m - 1 and np.linalg.det(cov_matrix) < 0:
        S[m - 1, m - 1] = -1
    elif cov_rank < m - 1:
        raise ValueError("colinearility detected in covariance matrix:\n{}".format(cov_matrix))

    R = U.dot(S).dot(V_t)
    c = (d * S.diagonal()).sum() / sigma_from
    t = mean_to - c * R.dot(mean_from)

    return c, R, t


def compute_rigid_transform(A, B):
    assert A.shape == B.shape

    num_rows, num_cols = A.shape
    if num_rows != 3:
        raise Exception(f"matrix A is not 3xN, it is {num_rows}x{num_cols}")

    num_rows, num_cols = B.shape
    if num_rows != 3:
        raise Exception(f"matrix B is not 3xN, it is {num_rows}x{num_cols}")

    # find mean column wise
    centroid_A = np.mean(A, axis=1)
    centroid_B = np.mean(B, axis=1)

    # ensure centroids are 3x1
    centroid_A = centroid_A.reshape(-1, 1)
    centroid_B = centroid_B.reshape(-1, 1)

    # subtract mean
    Am = A - centroid_A
    Bm = B - centroid_B

    H = Am @ np.transpose(Bm)

    # sanity check
    # if linalg.matrix_rank(H) < 3:
    #    raise ValueError("rank of H = {}, expecting 3".format(linalg.matrix_rank(H)))

    # find rotation
    U, S, Vt = np.linalg.svd(H)
    R = Vt.T @ U.T

    # special reflection case
    if np.linalg.det(R) < 0:
        print("det(R) < R, reflection detected!, correcting for it ...")
        Vt[2, :] *= -1
        R = Vt.T @ U.T

    t = -R @ centroid_A + centroid_B

    return R, np.squeeze(t)


def umeyama(from_points: np.ndarray, to_points: np.ndarray, estimate_scale: bool = False) -> np.ndarray:
    """Estimate N-D similarity transformation with or without scaling.
    See: https://github.com/iperov/DeepFaceLab/blob/master/core/mathlib/umeyama.py

    Parameters
    ----------
        from_points : (M, N) array
            Source coordinates.
        to_points : (M, N) array
            Destination coordinates.
        estimate_scale : bool
            Whether to estimate scaling factor.

    Returns
    -------
        T : (N + 1, N + 1)
            The homogeneous similarity transformation matrix. The matrix contains
            NaN values only if the problem is not well-conditioned.
            The matrix transforms from_points into to_points by: homogenize(from_points) @ T.T

    References
    ----------
        .. [1] "Least-squares estimation of transformation parameters between two
                point patterns", Shinji Umeyama, PAMI 1991, DOI: 10.1109/34.88573
    """

    num = from_points.shape[0]
    dim = from_points.shape[1]

    # Compute mean of src and dst.
    src_mean = from_points.mean(axis=0)
    dst_mean = to_points.mean(axis=0)

    # Subtract mean from src and dst.
    src_demean = from_points - src_mean
    dst_demean = to_points - dst_mean

    # Eq. (38).
    A = np.dot(dst_demean.T, src_demean) / num

    # Eq. (39).
    d = np.ones((dim,), dtype=np.double)
    if np.linalg.det(A) < 0:
        d[dim - 1] = -1

    T = np.eye(dim + 1, dtype=np.double)

    U, S, V = np.linalg.svd(A)

    # Eq. (40) and (43).
    rank = np.linalg.matrix_rank(A)
    if rank == 0:
        return np.nan * T
    elif rank == dim - 1:
        if np.linalg.det(U) * np.linalg.det(V) > 0:
            T[:dim, :dim] = np.dot(U, V)
        else:
            s = d[dim - 1]
            d[dim - 1] = -1
            T[:dim, :dim] = np.dot(U, np.dot(np.diag(d), V))
            d[dim - 1] = s
    else:
        T[:dim, :dim] = np.dot(U, np.dot(np.diag(d), V))

    if estimate_scale:
        # Eq. (41) and (42).
        scale = 1.0 / src_demean.var(axis=0).sum() * np.dot(S, d)
    else:
        scale = 1.0

    T[:dim, dim] = dst_mean - scale * np.dot(T[:dim, :dim], src_mean.T)
    T[:dim, :dim] *= scale

    return T



from .intrinsics_numpy import Intrinsics
from .intrinsics_torch import TorchIntrinsics
from .pose_numpy import Pose, PoseType, CameraCoordinateConvention
from .pose_torch import TorchPose
from .transform_numpy import compute_similarity_transform, compute_rigid_transform, umeyama



from typing import Optional, Union, Tuple

import numpy as np
import pyvista as pv

# TODO: Points is not a float type. This can cause issues when transforming or applying filters.
#  Casting to ``np.float32``. Disable this by passing ``force_float=False``
import vtk

from dreifus.camera import PoseType, CameraCoordinateConvention
from dreifus.matrix import Pose, Intrinsics
from dreifus.vector import Vec3, Vec4
from pyvista import examples


def add_coordinate_axes(p: pv.Plotter,
                        max_length: int = 10,
                        ticks_frequency: float = 1,
                        line_width: float = 1,
                        tick_length: float = 0.2,
                        arrow_size: float = 0.5,
                        draw_labels: bool = True,
                        scale: float = 1):
    axis_color_map = {
        0: 'r',
        1: 'g',
        2: 'b'
    }

    max_length *= scale
    ticks_frequency *= scale
    line_width *= scale
    tick_length *= scale
    arrow_size *= scale

    def _add_tick_lines(axis_across: int, axis_tick: int):
        points = []
        labels = []

        for pos in np.arange(-max_length, max_length, ticks_frequency):
            if pos == 0:
                continue

            point_1 = Vec3()
            point_1[axis_across] = pos
            point_1[axis_tick] = - tick_length

            point_2 = Vec3()
            point_2[axis_across] = pos
            point_2[axis_tick] = tick_length

            # # Apply Scaling
            # point_1 *= scale
            # point_2 *= scale

            p.add_lines(np.array([point_1, point_2]), color=axis_color_map[axis_across], width=line_width)

            points.append(point_2)
            labels.append(pos if round(pos) == pos else f"{pos:.1f}")

        if draw_labels:
            p.add_point_labels(points, labels, fill_shape=False, shape=None, show_points=False,
                               text_color=axis_color_map[axis_across])

    def _add_arrow(axis: int):
        cone_center = Vec3()
        cone_center[axis] = max_length - arrow_size / 2
        # cone_center *= scale
        p.add_mesh(pv.Cone(center=cone_center, direction=cone_center, height=arrow_size, radius=arrow_size / 2),
                   color=axis_color_map[axis])

    # Axes
    p.add_lines(np.array([[-max_length, 0, 0], [max_length - arrow_size, 0, 0]]), color='r', width=line_width)
    p.add_lines(np.array([[0, -max_length, 0], [0, max_length - arrow_size, 0]]), color='g', width=line_width)
    p.add_lines(np.array([[0, 0, -max_length], [0, 0, max_length - arrow_size]]), color='b', width=line_width)

    # Ticks
    # x: ticks across y dim
    # y: ticks across x dim
    # z: ticks across x dim
    _add_tick_lines(0, 1)
    _add_tick_lines(1, 0)
    _add_tick_lines(2, 0)

    # Arrows
    _add_arrow(0)
    _add_arrow(1)
    _add_arrow(2)


def add_floor(p: pv.Plotter,
              square_size: float = 1,
              max_distance: float = 10,
              color='gray',
              axes: Tuple[int, int] = (0, 1)):
    idx1, idx2 = axes
    for i in np.arange(-max_distance, max_distance, square_size):
        for j in np.arange(-max_distance, max_distance, square_size):
            point_1 = Vec3()
            point_1[idx1] = i
            point_1[idx2] = j

            point_2 = Vec3()
            point_2[idx1] = i + square_size
            point_2[idx2] = j

            point_3 = Vec3()
            point_3[idx1] = i + square_size
            point_3[idx2] = j + square_size

            point_4 = Vec3()
            point_4[idx1] = i
            point_4[idx2] = j + square_size

            floor_square = pv.Rectangle([
                point_1, point_2, point_3, point_4
            ])
            # floor_square = pv.Rectangle([
            #     [x, y, 0],
            #     [x + square_size, y, 0],
            #     [x + square_size, y + square_size, 0],
            #     [x, y + square_size, 0]]
            # )
            p.add_mesh(floor_square, show_edges=True, opacity=0.2, color=color)


def add_camera_frustum(p: pv.Plotter,
                       pose: Pose,
                       intrinsics: Intrinsics,
                       img_w: Optional[float] = None,
                       img_h: Optional[float] = None,
                       image: Optional[np.ndarray] = None,
                       color='lightgray',
                       size: float = 0.3,
                       label: Optional[Union[str, int]] = None,
                       line_width: float = 1,
                       look_vector_length: float = 0):
    """
    Add a camera frustum visualizing the camera with the given `pose` and `intrinsics` to the pyvista plotter `p`.

    Parameters
    ----------
        p:
            the pyvista Plotter() object
        pose:
            the camera pose to visualize. A frustum will be drawn at the camera's world space position
        intrinsics:
            intrinsics of the camera. Necessary to properly visualize the viewing angle of the frustum
        img_w:
            Optionally, the width of the images that this camera produces. If not specified, it is assumed that
            the image width is 2 * intrinsics.cx
        img_h:
            Optionally, the height of the images that this camera produces. If not specified, it is assumed that
            the image height is 2 * intrinsics.cy
        image:
            Optionally, a color image that is visualized inside the camera frustum. Helpful for debugging camera
            locations multi-view image scenarios.
            Expected to be np.uint8
        color:
            Optionally, a color for the lines of the drawn camera frustum
        size:
            Size of the drawn camera frustum in depth units (camera center location to frustum pane)
        label:
            Optionally, a label for the frustum that will be drawn above
        line_width:
            The width of the frustum's lines
        look_vector_length:
            If >0, an additional line will be drawn from the camera location in look direction with the given length
            in depth units
    """

    if pose.pose_type == PoseType.WORLD_2_CAM:
        pose = pose.invert()

    pose = pose.change_camera_coordinate_convention(CameraCoordinateConvention.OPEN_CV, inplace=False)

    assert pose.pose_type == PoseType.CAM_2_WORLD
    assert pose.camera_coordinate_convention == CameraCoordinateConvention.OPEN_CV

    center = pose.get_translation()

    depth = size * pose.camera_coordinate_convention.forward_direction.sign()

    # Infer image width/height if it is not specifically provided
    if img_h is None:
        img_h = 2 * intrinsics.cy if image is None else image.shape[0]

    if img_w is None:
        img_w = 2 * intrinsics.cx if image is None else image.shape[1]

    # img_w = 2 * intrinsics.cx if img_w is None else img_w
    # img_h = 2 * intrinsics.cy if img_h is None else img_h

    # Assume that x -> right, y -> down
    p_top_left = Vec4(0, 0, depth, 1)
    p_top_right = Vec4(img_w * depth, 0, depth, 1)
    p_bottom_left = Vec4(0, img_h * depth, depth, 1)
    p_bottom_right = Vec4(img_w * depth, img_h * depth, depth, 1)

    # Triangle visualizes -y direction of cam space
    cam_y_sign = pose.camera_coordinate_convention.up_direction.sign()
    p_tri_left = Vec4(0, cam_y_sign * img_h * depth / 5, depth, 1)
    p_tri_right = Vec4(img_w * depth, cam_y_sign * img_h * depth / 5, depth, 1)
    p_tri_top = Vec4(img_w * depth / 2, cam_y_sign * img_h * depth / 2, depth, 1)

    points = np.stack([p_top_left, p_top_right, p_bottom_right, p_bottom_left, p_tri_left, p_tri_right, p_tri_top])

    # TODO: Not sure if intrinsics is used correctly here
    points_world = (pose @ intrinsics.homogenize(invert=True) @ points.T).T
    points_world_frustum = points_world[:4]
    points_world_up_triangle = points_world[4:]

    if image is not None:
        image_rectangle = pv.Quadrilateral([points_world_frustum[0][:3],
                                            points_world_frustum[1][:3],
                                            points_world_frustum[2][:3],
                                            points_world_frustum[3][:3]])
        # image_file = examples.mapfile
        # tex = pv.read_texture(image_file)
        if image.dtype != np.uint8:
            if image.max() < 2:
                # Assume that image is float, and we have to scale from [0, 1] -> [0, 255]
                image = (image * 255).clip(0, 255).astype(np.uint8)
        tex = pv.numpy_to_texture(image)
        image_rectangle.active_t_coords = np.array([[0, 1], [1, 1], [1, 0], [0, 0]])
        p.add_mesh(image_rectangle, texture=tex)

    # Draw lines from rectangle corners to camera origin
    for i_point, point_world in enumerate(points_world_frustum):
        next_point_world = points_world_frustum[(i_point + 1) % 4]
        p.add_lines(np.array([center, point_world[:3]]), width=line_width, color=color)
        p.add_lines(np.array([point_world[:3], next_point_world[:3]]), width=line_width, color=color)

    # Draw triangle
    p.add_lines(np.concatenate(
        [points_world_up_triangle[[0], :3],
         points_world_up_triangle[[1], :3],
         points_world_up_triangle[[1], :3],
         points_world_up_triangle[[2], :3],
         points_world_up_triangle[[2], :3],
         points_world_up_triangle[[0], :3]], axis=0),
        width=line_width,
        color=color)

    # Draw look direction line
    if look_vector_length > 0:
        position = pose.get_translation()
        look_direction = pose.get_look_direction()
        p.add_lines(np.stack([position, position + look_vector_length * look_direction]), width=line_width, color=color)

    if label is not None:
        p_center = center
        # (p_top_left + 0.5 * (p_bottom_right - p_top_left))[:3]
        p.add_point_labels([p_center], [label],
                           fill_shape=False, shape=None, show_points=False, text_color=color)


def add_coordinate_system(p: pv.Plotter, cam_to_world: Pose, scale: float = 1, line_width: float = 1):
    axis_color_map = {
        0: 'r',
        1: 'g',
        2: 'b'
    }

    origin = cam_to_world.get_translation()
    px = Vec4(scale, 0, 0, 1)
    py = Vec4(0, scale, 0, 1)
    pz = Vec4(0, 0, scale, 1)

    points = np.stack([px, py, pz])
    points_world = (cam_to_world @ points.T).T

    for i_point, point_world in enumerate(points_world):
        color = axis_color_map[i_point]
        p.add_lines(np.array([origin, point_world[:3]]), width=line_width, color=color)


def set_camera(p: pv.Plotter, cam_to_world: Pose, neg_z_forward: bool = False):
    look_direction = cam_to_world.get_look_direction()
    up_direction = cam_to_world.get_up_direction()
    if neg_z_forward:
        look_direction *= -1
        # up_direction *= -1
    focal_point = cam_to_world.get_translation() + look_direction

    p.camera_set = True
    p.camera_position = cam_to_world.get_translation().tolist()
    p.camera.focal_point = focal_point
    p.camera.up = up_direction  # TODO: Do we need to negate here? OpenCV coordinate y goes down


def render_from_camera(p: pv.Plotter,
                       pose: Pose,
                       intrinsics: Intrinsics) -> np.ndarray:
    """
    Render the given scene from the specified camera.
    The image size will be the size of the pyvista window.
    The pyvista plotter should be initialized like this:
    ```
    p = pv.Plotter(off_screen=True, window_size=[IMG_W, IMG_H])
    ```

    Parameters
    ----------
        p: the pyvista scene to render
        pose: extrinsics camera pose
        intrinsics: camera intrinsics

    Returns
    -------
        A numpy array [IMG_H, IMG_W, 4] containing the rendered scene. The background is transparent and a mask can
        be obtained from the alpha channel
    """

    pose = pose.change_pose_type(PoseType.CAM_2_WORLD, inplace=False)
    pose = pose.change_camera_coordinate_convention(CameraCoordinateConvention.OPEN_CV, inplace=False)
    pose = pose.change_pose_type(PoseType.WORLD_2_CAM, inplace=False)

    w = p.window_size[0]
    h = p.window_size[1]

    # ----------------------------------------------------------
    # Intrinsics
    # ----------------------------------------------------------
    cx = intrinsics.cx
    cy = intrinsics.cy
    fx = intrinsics.fx
    fy = intrinsics.fy

    # convert the principal point to window center (normalized coordinate system) and set it
    wcx = -2 * (cx - float(w) / 2) / w
    wcy = 2 * (cy - float(h) / 2) / h
    p.camera.SetWindowCenter(wcx, wcy)

    # convert the focal length to view angle and set it
    view_angle = 180 / np.pi * (2.0 * np.arctan2(h / 2.0, fx))
    p.camera.SetViewAngle(view_angle)

    # ----------------------------------------------------------
    # Extrinsics
    # ----------------------------------------------------------

    # apply the transform to scene objects
    vtk_pose = vtk.vtkMatrix4x4()
    for i in range(pose.shape[0]):
        for j in range(pose.shape[1]):
            vtk_pose.SetElement(i, j, pose[i, j])

    p.camera.SetModelTransformMatrix(vtk_pose)

    # the camera can stay at the origin because we are transforming the scene objects
    p.camera.SetPosition(0, 0, 0)

    # look in the +Z direction of the camera coordinate system
    p.camera.SetFocalPoint(0, 0, 1)

    # the camera Y axis points down
    p.camera.SetViewUp(0, -1, 0)

    # ensure the relevant range of depths are rendered
    p.renderer.ResetCameraClippingRange()

    p.render()  # Important, otherwise view isn't updated when render_from_camera() is called multiple times
    img = np.asarray(p.screenshot(transparent_background=True))

    return img



from typing import Optional, Union, List, Tuple

import cv2
import numpy as np
from scipy.spatial.transform import Rotation as R

from dreifus.camera import CameraCoordinateConvention, PoseType
from dreifus.matrix.pose_base import is_rotation_matrix
from dreifus.vector.vector_base import Vec3TypeX, FloatType, unpack_3d_params, Vec3Type, Vec4Type
from dreifus.vector.vector_numpy import Vec3, Vec4


class Pose(np.ndarray):
    camera_coordinate_convention: CameraCoordinateConvention
    pose_type: PoseType
    _disable_rotation_check: bool = False

    def __new__(cls,
                matrix_or_rotation: Union[np.ndarray, List] = np.eye(4),
                translation: Optional[Vec3Type] = None,
                camera_coordinate_convention: CameraCoordinateConvention = CameraCoordinateConvention.OPEN_CV,
                pose_type: PoseType = PoseType.WORLD_2_CAM,
                disable_rotation_check: bool = False):
        """
        Per default, Poses are assumed to be WORLD_2_CAM poses that transform world coordinates into an OPEN_CV
        camera space.

        Parameters
        ----------
            matrix_or_rotation:
                Either a full 4x4 affine transform matrix or a 3x3 rotation matrix
            translation:
                If `matrix_or_rotation` was only a 3x3 rotation matrix, ´translation` can be specified as a 3x1 vector
            camera_coordinate_convention:
                Indicates which coordinate convention is used for the camera space that this pose matrix maps into
            pose_type:
                Indicates whether this pose maps from WORLD -> CAM or CAM -> WORLD (inverse)
        """

        pose = super().__new__(cls, (4, 4), dtype=np.float32)

        pose.camera_coordinate_convention = camera_coordinate_convention
        pose.pose_type = pose_type
        pose._disable_rotation_check = disable_rotation_check

        if not isinstance(matrix_or_rotation, np.ndarray):
            matrix_or_rotation = np.asarray(matrix_or_rotation)

        if matrix_or_rotation.shape == (4, 4):
            # Full 4x4 Pose
            assert translation is None, "If a full pose is given, no translation should be specified!"
            assert (matrix_or_rotation[3, :] == [0, 0, 0, 1]).all(), \
                f"Last row of pose must be [0, 0, 0, 1]. Got {matrix_or_rotation[3, :]}"
            assert disable_rotation_check or is_rotation_matrix(matrix_or_rotation[:3, :3]), \
                f"Specified matrix does not contain a valid rotation matrix! {np.array(matrix_or_rotation[:3, :3])}"

            pose[:] = matrix_or_rotation

        elif matrix_or_rotation.shape == (3, 3):
            # 3x3 rotation matrix + 3(x1) translation vector
            translation = np.asarray(translation)
            assert translation.squeeze().shape == (3,), \
                "If a rotation matrix is given, the translation vector has to be 3 dimensional!"

            pose[:3, :3] = matrix_or_rotation
            pose[:3, 3] = translation.squeeze()
            pose[3, :] = [0, 0, 0, 1]
        elif matrix_or_rotation.shape == (3, 1) or matrix_or_rotation.shape == (3,):
            # 3(x1) Rodriguez vector + 3(x1) translation vector
            assert translation.squeeze().shape == (3,), \
                "If a Rodriguez vector is given, the translation vector has to be 3 dimensional!"

            pose[:3, :3] = cv2.Rodrigues(matrix_or_rotation)[0]
            pose[:3, 3] = translation.squeeze()
            pose[3, :] = [0, 0, 0, 1]
        else:
            raise ValueError("Either a full pose has to be given or a 3x3 rotation + 3x1 translation!")

        return pose

    @staticmethod
    def from_euler(euler_angles: Vec3Type,
                   translation: Vec3Type = Vec3(),
                   euler_mode: str = 'XYZ',
                   camera_coordinate_convention: CameraCoordinateConvention = CameraCoordinateConvention.OPEN_CV,
                   pose_type: PoseType = PoseType.WORLD_2_CAM) -> 'Pose':
        return Pose(R.from_euler(euler_mode, euler_angles).as_matrix(),
                    translation,
                    camera_coordinate_convention=camera_coordinate_convention,
                    pose_type=pose_type)

    @staticmethod
    def from_rodriguez(rodriguez_vector: Vec3Type,
                       translation: Vec3Type = Vec3(),
                       camera_coordinate_convention: CameraCoordinateConvention = CameraCoordinateConvention.OPEN_CV,
                       pose_type: PoseType = PoseType.WORLD_2_CAM) -> 'Pose':
        return Pose(cv2.Rodrigues(rodriguez_vector)[0],
                    translation,
                    camera_coordinate_convention=camera_coordinate_convention,
                    pose_type=pose_type)

    @staticmethod
    def from_quaternion(quaternion: Vec4Type,
                        translation: Vec3Type = Vec3(),
                        camera_coordinate_convention: CameraCoordinateConvention = CameraCoordinateConvention.OPEN_CV,
                        pose_type: PoseType = PoseType.WORLD_2_CAM) -> 'Pose':
        return Pose(R.from_quat(quaternion).as_matrix(),
                    translation,
                    camera_coordinate_convention=camera_coordinate_convention,
                    pose_type=pose_type)

    def get_rotation_matrix(self) -> np.ndarray:
        return self[:3, :3]

    def get_euler_angles(self, order: str) -> Vec3:
        return Vec3(R.from_matrix(self.get_rotation_matrix()).as_euler(order))

    def get_rodriguez_vector(self) -> Vec3:
        return Vec3(cv2.Rodrigues(self.get_rotation_matrix())[0].squeeze())

    def get_quaternion(self) -> Vec4:
        return Vec4(R.from_matrix(self.get_rotation_matrix()).as_quat())

    def get_translation(self) -> Vec3:
        # assert self.pose_type == PoseType.CAM_2_WORLD, "camera position only makes sense for CAM_2_WORLD poses"
        return Vec3(self[:3, 3])

    def set_translation(self, x: Optional[Vec3TypeX] = None, y: Optional[FloatType] = None, z: Optional[FloatType] = None):
        x, y, z = unpack_3d_params(x, y, z)
        if x is not None:
            self[0, 3] = x
        if y is not None:
            self[1, 3] = y
        if z is not None:
            self[2, 3] = z

    def move(self, x: Optional[Vec3TypeX] = None, y: Optional[FloatType] = None, z: Optional[FloatType] = None,
             inplace: bool = True) -> 'Pose':
        x, y, z = unpack_3d_params(x, y, z, 0)

        if inplace:
            pose = self
        else:
            pose = self.copy()

        pose[0, 3] += x
        pose[1, 3] += y
        pose[2, 3] += z

        return pose

    def scale(self, scale: float) -> 'Pose':
        self[:3, 3] *= scale
        return self

    def set_rotation_matrix(self, rotation_matrix: np.ndarray):
        assert self._disable_rotation_check or is_rotation_matrix(rotation_matrix), \
            f"Specified matrix does not contain a valid rotation matrix! {rotation_matrix}"
        self[:3, :3] = rotation_matrix

    def set_rotation_euler(self, order: str,
                           euler_x: Vec3TypeX = 0,
                           euler_y: Optional[float] = None,
                           euler_z: Optional[float] = None):
        euler_x, euler_y, euler_z = unpack_3d_params(euler_x, euler_y, euler_z, default=0)
        self[:3, :3] = R.from_euler(order, [euler_x, euler_y, euler_z]).as_matrix()

    def rotate_euler(self,
                     order: str,
                     euler_x: Vec3TypeX = 0,
                     euler_y: Optional[float] = None,
                     euler_z: Optional[float] = None,
                     inplace: bool = True) -> 'Pose':
        if inplace:
            pose = self
        else:
            pose = self.copy()

        euler_x, euler_y, euler_z = unpack_3d_params(euler_x, euler_y, euler_z, default=0)
        euler_rotation = Vec3(euler_x, euler_y, euler_z)
        current_euler_angles = pose.get_euler_angles(order)
        pose.set_rotation_euler(order, current_euler_angles + euler_rotation)

        return pose

    def invert(self, inplace: bool = False) -> 'Pose':
        assert not self._disable_rotation_check, "Cannot invert when no proper rotation matrix is used"
        inverted_rotation = self.get_rotation_matrix().T
        inverted_translation = -inverted_rotation @ self.get_translation()

        if inplace:
            self.set_rotation_matrix(inverted_rotation)
            self.set_translation(inverted_translation)
            self.pose_type = self.pose_type.invert()
            inverted_pose = self
        else:
            inverted_pose = Pose(inverted_rotation,
                                 inverted_translation,
                                 camera_coordinate_convention=self.camera_coordinate_convention,
                                 pose_type=self.pose_type.invert())
        return inverted_pose

    def negate_orientation_axis(self, axis: int):
        """
        Changes the coordinate convention of the camera space.
        E.g., if the pose is assumed to be in OpenCV camera space (x -> right, y -> down, z -> forward), then
        negate_orientation_axis(1) will make it DirectX (x -> right, y -> up, z -> forward).
        The function negates the column of the rotation matrix that corresponds to the given axis.
        The camera's position in world space is NOT affected by this transformation.
        Only the orientation of the camera frustum is changed.
        Assumes the current pose is cam2world.

        Parameters
        ----------
            axis: which axis to negate.
        """

        # negates the column of the rotation matrix
        self[:3, axis] *= -1

    def swap_axes(self, permutation: List[Union[int, str]], inplace: bool = True) -> 'Pose':
        """
        Negates/swaps entire rows of the pose matrix.
        Essentially, applies a rotation / flip operation around the world origin to the camera object.
        E.g., swap_axes(['-x', 'y', 'z']) will move an origin-facing camera that was at (3, 1, 1) to (-3, 1, 1).
        The moved camera will still face the origin afterwards (i.e., the camera is not just moved but also rotated).
        Alternatively, instead of flipping the camera along axis-aligned planes, swap_axes() can be interpreted as
        flipping the actual scene in world space while keeping the camera where it was.
        I.e., after applying the camera transformation the scene will be rendered as if it was flipped.
        Assumes the current pose is cam2world

        Parameters
        ----------
            permutation: 3-tuple of axis indicators (either 0, 1, 2 or x, y, z with optional '-' signs)
        """

        if inplace:
            pose = self
        else:
            pose = self.copy()

        axis_switcher = np.zeros((4, 4))
        axis_order = ['x', 'y', 'z']
        for idx, a in enumerate(permutation):
            v = 1
            if isinstance(a, int):
                ax = a
            else:
                # Possibility to also flip an axis via -x, -y etc.
                ax = axis_order.index(a[-1])  # Map x -> 0, y -> 1, z -> 2
                if a[0] == '-':
                    # Axis shall be flipped
                    v = -1

            axis_switcher[idx, ax] = v
        axis_switcher[3, 3] = 1

        if np.abs(np.linalg.det(axis_switcher) - 1) > 1e-6:
            print("[WARNING] swap_axis changes handedness!")

        # Negates / Flips rows
        pose[:, :] = axis_switcher @ pose

        return pose

    def change_camera_coordinate_convention(self,
                                            new_camera_coordinate_convention: CameraCoordinateConvention,
                                            inplace: bool = True) -> 'Pose':
        # TODO: Make this work for WORLD_2_CAM poses as well
        assert self.pose_type == PoseType.CAM_2_WORLD, "Camera coordinate conventions can only be changed on CAM_2_WORLD matrices"

        current_ccc = self.camera_coordinate_convention

        if inplace:
            pose = self
        else:
            pose = self.copy()

        if current_ccc.x_direction != new_camera_coordinate_convention.x_direction:
            pose.negate_orientation_axis(0)

        if current_ccc.y_direction != new_camera_coordinate_convention.y_direction:
            pose.negate_orientation_axis(1)

        if current_ccc.z_direction != new_camera_coordinate_convention.z_direction:
            pose.negate_orientation_axis(2)

        pose.camera_coordinate_convention = new_camera_coordinate_convention

        return pose

    def change_pose_type(self,
                         new_pose_type: PoseType,
                         inplace: bool = True) -> 'Pose':
        assert self.pose_type in {PoseType.CAM_2_WORLD, PoseType.WORLD_2_CAM}, \
            "start pose must be either cam2world or world2cam"
        assert new_pose_type in {PoseType.CAM_2_WORLD, PoseType.WORLD_2_CAM}, \
            "target pose type must be either cam2world or world2cam"

        if inplace:
            pose = self
        else:
            pose = self.copy()

        if pose.pose_type != new_pose_type:
            pose = pose.invert(inplace=inplace)

        return pose

    def get_look_direction(self) -> 'Vec3':
        # Assumes the current pose is cam2world
        # Assigns meaning to the coordinate axes. Hence, the coordinate system convention is important
        # Assumes an OpenCV camera coordinate system convention (x -> right, y -> down, z -> forward/look)

        assert self.pose_type == PoseType.CAM_2_WORLD
        forward_direction = self.camera_coordinate_convention.forward_direction
        axis = forward_direction.axis_id
        sign = forward_direction.sign()
        look_direction = sign * Vec3(self[:3, axis])

        return look_direction

    def get_up_direction(self) -> 'Vec3':
        # Assumes the current pose is cam2world
        # Assigns meaning to the coordinate axes. Hence, the coordinate system convention is important
        # Assumes an OpenCV camera coordinate system convention (x -> right, y -> down, z -> forward)

        assert self.pose_type == PoseType.CAM_2_WORLD
        up_direction = self.camera_coordinate_convention.up_direction
        axis = up_direction.axis_id
        sign = up_direction.sign()

        up_direction = sign * Vec3(self[:3, axis])

        return up_direction

    def look_at(self, at: Vec3, up: Vec3 = Vec3(0, 0, 1)):
        # This method assigns meaning to the coordinate axes. Hence, the coordinate system convention is important
        # We use the OpenCV camera coordinate system convention

        # Poses are always assumed to be cam2world
        # That way the translation part of the pose matrix is the location of the object in world space

        assert self.pose_type == PoseType.CAM_2_WORLD
        assert self.camera_coordinate_convention == CameraCoordinateConvention.OPEN_CV

        eye = self.get_translation()
        z_axis = (at - eye).normalize()  # Assumes z-axis is forward
        x_axis = z_axis.cross(up).normalize()  # Assumes y-axis is up
        y_axis = x_axis.cross(z_axis).normalize()

        # Important as otherwise rotation matrix has negative determinant (would be left-handed).
        # Makes it a [x, -y, z] OpenCV camera coordinate system
        # [x, y, -z] would be a Blender/OpenGL camera coordinate system
        y_axis = - y_axis

        self.set_rotation_matrix(np.array([x_axis, y_axis, z_axis]).T)
        # self.set_translation(np.dot(x_axis, eye), np.dot(y_axis, eye), np.dot(z_axis, eye))
        self.set_translation(eye)

    def calculate_rigid_transformation_to(self, other: 'Pose') -> 'Pose':
        """
        Let current cam2world pose = X and target cam2world pose = Y.
        Calculate a rigid transform T, s.t. T @ X = Y
        In other words, T shall rotate and move the coordinate system of X such that it aligns with the coordinate
        system of Y.

        Parameters
        ----------
            other:
                the target pose. Must be either cam2world or world2cam

        Returns
        -------
            The rigid transformation T (a cam2cam pose) that aligns the current pose with the specified pose.
        """

        assert self.pose_type in {PoseType.CAM_2_WORLD, PoseType.WORLD_2_CAM}, \
            "start pose must be either cam2world or world2cam"
        assert other.pose_type in {PoseType.CAM_2_WORLD, PoseType.WORLD_2_CAM}, \
            "target pose must be either cam2world or world2cam"

        # TODO: Continue
        source_cam2world = self.change_pose_type(PoseType.CAM_2_WORLD, inplace=False)
        target_cam2world = other.change_pose_type(PoseType.CAM_2_WORLD, inplace=False)

        rigid_transformation = target_cam2world @ source_cam2world.invert()
        assert rigid_transformation.pose_type == PoseType.CAM_2_CAM

        return rigid_transformation

    def __rmatmul__(self, other):
        if isinstance(other, Pose):
            pose = super(Pose, self).__rmatmul__(other)
            pose.camera_coordinate_convention = self.camera_coordinate_convention
            pose.pose_type = self.pose_type @ other.pose_type
            return pose
        else:
            return other @ np.array(self)

    def __matmul__(self, other):
        # TODO: figure out why numpy operations automatically cast to Pose again
        if isinstance(other, Pose):
            pose = super(Pose, self).__matmul__(other)
            pose.camera_coordinate_convention = self.camera_coordinate_convention
            pose.pose_type = self.pose_type @ other.pose_type
            return pose
        else:
            return np.array(self) @ other

    def copy(self, order='C') -> 'Pose':
        pose = super(Pose, self).copy(order=order)
        pose.camera_coordinate_convention = self.camera_coordinate_convention
        pose.pose_type = self.pose_type

        return pose

    def numpy(self) -> np.ndarray:
        return np.array(self)

    def __repr__(self):
        representation = super(Pose, self).__repr__()
        representation = f"{representation}\n" \
                         f" > camera_coordinate_convention: {self.camera_coordinate_convention.name}\n" \
                         f" > pose_type: {self.pose_type.name}"
        return representation

    def __str__(self):
        string = super(Pose, self).__str__()
        string = f"{string}\n" \
                 f" > camera_coordinate_convention: {self.camera_coordinate_convention.name}\n" \
                 f" > pose_type: {self.pose_type.name}"
        return string



import math
from typing import Optional, Tuple, Union, List

import numpy as np

# ==========================================================
# 3D utilities
# ==========================================================
from scipy.linalg import expm, norm

from dreifus.vector.vector_base import Vec3TypeX, unpack_3d_params, unpack_nd_params


def rotation_matrix_between_vectors(vec1: Vec3TypeX, vec2: Vec3TypeX) -> np.ndarray:
    """ Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    vec1 = Vec3(vec1)
    vec2 = Vec3(vec2)

    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    if any(v):  # if not all zeros then
        c = np.dot(a, b)
        s = np.linalg.norm(v)
        kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        return np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))

    else:
        return np.eye(3)  # cross of all zeros only occurs on identical directions


def rotation_matrix_around_axis(axis: Vec3TypeX, theta: float) -> np.ndarray:
    """
    Computes a 3x3 rotation matrix that moves 3D points in space around the specified axis by theta radians.

    Parameters
    ----------
        axis: Which axis to rotate around (always goes through origin)
        theta: How much to rotate in radians. Rotation is applied in clockwise fashion (right-hand rotation)

    Returns
    -------
        The 3x3 rotation matrix
    """

    return expm(np.cross(np.eye(3), axis / norm(axis) * theta))


def angle_between_vectors(vec1: Vec3TypeX, vec2: Vec3TypeX) -> float:
    vec1 = Vec3(vec1)
    vec2 = Vec3(vec2)
    return math.acos(vec1.dot(vec2) / (vec1.length() * vec2.length()))


def offset_vector_between_line_and_point(offset: Vec3TypeX, direction: Vec3TypeX, point: Vec3TypeX):
    offset = Vec3(offset)
    direction = Vec3(direction).normalize()
    point = Vec3(point)

    direction_to_point = point - offset
    closest_point_on_line = offset + direction_to_point.dot(direction) * direction
    move = point - closest_point_on_line
    return move


def to_homogeneous(points: np.ndarray) -> np.ndarray:
    ones = np.ones((*points.shape[:-1], 1), dtype=points.dtype)
    points = np.concatenate([points, ones], axis=-1)

    return points


# ==========================================================
# Numpy Vectors
# ==========================================================

class Vec2(np.ndarray):
    def __new__(cls,
                x: Optional[Union[float, Tuple[float, float], List[float], np.ndarray]] = None,
                y: Optional[float] = None) -> 'Vec2':
        vec2 = super().__new__(cls, (2,), dtype=np.float32)
        x, y = unpack_nd_params(2, x, y, default=0)
        # try:
        #     assert len(x) == 2, "Passed list must contain exactly 3 values"
        #     assert y is None, "When a list is passed, y must not be given"
        #     x, y = x
        # except TypeError:
        #     # x was actually a number
        #     pass
        vec2[:] = [x, y]
        return vec2

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @x.setter
    def x(self, x: float):
        self[0] = x

    @y.setter
    def y(self, y: float):
        self[1] = y

    def length(self) -> float:
        return np.linalg.norm(self)

    def normalize(self) -> 'Vec2':
        return self / self.length()


class Vec3(np.ndarray):
    def __new__(cls,
                x: Optional[Vec3TypeX] = None,
                y: Optional[float] = None,
                z: Optional[float] = None) -> 'Vec3':
        # TODO: Allow no arguments -> 0 vector
        vec3 = super().__new__(cls, (3,), dtype=np.float32)
        x, y, z = unpack_3d_params(x, y, z, default=0)
        vec3[:] = [x, y, z]
        return vec3

    @staticmethod
    def unit(dimension: int) -> 'Vec3':
        unit_vector = Vec3()
        unit_vector[dimension] = 1
        return unit_vector

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @x.setter
    def x(self, x: float):
        self[0] = x

    @y.setter
    def y(self, y: float):
        self[1] = y

    @z.setter
    def z(self, z: float):
        self[2] = z

    def length(self) -> float:
        return np.linalg.norm(self)

    def normalize(self) -> 'Vec3':
        return self / self.length()

    def cross(self, other: 'Vec3') -> 'Vec3':
        return Vec3(np.cross(self, other))

    def homogenize(self) -> 'Vec4':
        return Vec4(self, 1)

    def sum(self) -> float:
        # Ensure that .sum() does not return a Vec3 object with just one value
        return super(Vec3, self).sum().item()

    def __eq__(self, other) -> bool:
        assert isinstance(other, Vec3), "Can only compare against other Vec3 instances"

        return (self.x == other.x and self.y == other.y and self.z == other.z)


class Vec4(np.ndarray):
    def __new__(cls,
                x: Union[float, Tuple[float, float, float, float], np.ndarray, 'Vec4', 'Vec3'],
                y: Optional[float] = None,
                z: Optional[float] = None,
                w: Optional[float] = None) -> 'Vec4':
        vec4 = super().__new__(cls, (4,), dtype=np.float32)
        x, y, z, w = unpack_nd_params(4, x, y, z, w)
        vec4[:] = [x, y, z, w]
        return vec4

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @property
    def w(self) -> float:
        return self[3]

    @x.setter
    def x(self, x: float):
        self[0] = x

    @y.setter
    def y(self, y: float):
        self[1] = y

    @z.setter
    def z(self, z: float):
        self[2] = z

    @w.setter
    def w(self, w: float):
        self[3] = w

    def length(self) -> float:
        return np.linalg.norm(self)

    def normalize(self) -> 'Vec4':
        return self / self.length()
